#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Metadata constants synopsizing high-level application behaviour.

Python Version
----------
For uniformity between this codebase and the ``setup.py`` setuptools script
importing this module, this module also validates the version of the active
Python 3 interpreter. An exception is raised if this version is insufficient.

This application currently requires **Python 3.4**, as Python < 3.4:

* Provides insufficient machinery for dynamically inspecting modules at runtime.
  In particular, both the long-standing :func:`imp.find_module` function and the
  :func:`importlib.find_loader` function introduced by Python 3.3 require all
  parent packages of the passed module to be recursively imported *before* these
  functions are called; failing to do so results in these functions
  unconditionally returning ``None``. Since this has been the source of numerous
  subtle issues throughout this codebase, Python 3.3 is strictly out. Since most
  modern Linux distributions have adopted Python 3.4 as the default Python
  3 interpreters, this *should* impose no hardship.
* Fails to provide the :mod:`enum` module introduced by Python 3.4, which both
  standardizes and simplifies enumeration implementations.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from modules guaranteed to exist at the start of
# installation. This includes all standard Python and application modules but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import sys

# ....................{ METADATA                           }....................
NAME = 'BETSE'
'''
Human-readable application name.
'''


LICENSE = '2-clause BSD'
'''
Human-readable name of the license this application is licensed under.
'''

# ....................{ PYTHON ~ version                   }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: Changes to this section *MUST* be synchronized with:
# * The corresponding section of the "betsee.metadata" submodule.
# * Front-facing documentation (e.g., "README.rst", "doc/md/INSTALL.md").
# On bumping the minimum required version of Python, consider also documenting
# the justification for doing so in the "Python Version" section of this
# submodule's docstring above.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

PYTHON_VERSION_MIN = '3.4.0'
'''
Human-readable minimum version of Python required by this application as a
``.``-delimited string.
'''


def _convert_version_str_to_tuple(version_str: str) -> tuple:
    '''
    Convert the passed human-readable ``.``-delimited version string into a
    machine-readable version tuple of corresponding integers.
    '''
    assert isinstance(version_str, str), (
        '"{}" not a version string.'.format(version_str))

    return tuple(int(version_part) for version_part in version_str.split('.'))


PYTHON_VERSION_MIN_PARTS = _convert_version_str_to_tuple(PYTHON_VERSION_MIN)
'''
Machine-readable minimum version of Python required by this application as a
tuple of integers.
'''


# Validate the version of the active Python interpreter *BEFORE* subsequent code
# possibly depending on such version. Since such version should be validated
# both at setuptools-based install time and post-install runtime *AND* since
# this module is imported sufficiently early by both, stash such validation here
# to avoid duplication of such logic and hence the hardcoded Python version.
#
# The "sys" module exposes three version-related constants for this purpose:
#
# * "hexversion", an integer intended to be specified in an obscure (albeit
#   both efficient and dependable) hexadecimal format: e.g.,
#    >>> sys.hexversion
#    33883376
#    >>> '%x' % sys.hexversion
#    '20504f0'
# * "version", a human-readable string: e.g.,
#    >>> sys.version
#    2.5.2 (r252:60911, Jul 31 2008, 17:28:52)
#    [GCC 4.2.3 (Ubuntu 4.2.3-2ubuntu7)]
# * "version_info", a tuple of three or more integers *OR* strings: e.g.,
#    >>> sys.version_info
#    (2, 5, 2, 'final', 0)
#
# For sanity, this application will *NEVER* conditionally depend upon the
# string-formatted release type of the current Python version exposed via the
# fourth element of the "version_info" tuple. Since the first three elements of
# that tuple are guaranteed to be integers *AND* since a comparable 3-tuple of
# integers is declared above, comparing the former and latter yield the simplest
# and most reliable Python version test.
#
# Note that the nearly decade-old and officially accepted PEP 345 proposed a new
# field "requires_python" configured via a key-value pair passed to the call to
# setup() in "setup.py" (e.g., "requires_python = ['>=2.2.1'],"), that field has
# yet to be integrated into either disutils or setuputils. Hence, that field is
# validated manually in the typical way. Behead the infidel setuptools!
if sys.version_info[:3] < PYTHON_VERSION_MIN_PARTS:
    # Human-readable current version of Python. "sys.version" is sufficiently
    # overly verbose as to be unusuable, sadly.
    PYTHON_VERSION = '.'.join(
        str(version_part) for version_part in sys.version_info[:3])

    # Die ignominiously.
    raise RuntimeError(
        '{} requires at least Python {}, but the active Python interpreter '
        'is only Python {}. We feel deep sadness for you.'.format(
            NAME, PYTHON_VERSION_MIN, PYTHON_VERSION))

# ....................{ METADATA ~ version                 }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: When modifying the current version of this application below,
# consider adhering to the Semantic Versioning schema. Specifically, the version
# should consist of three "."-delimited integers "{major}.{minor}.{patch}",
# where:
#
# * "{major}" specifies the major version, incremented only when either:
#   * Breaking configuration file backward compatibility. Since this
#     application's public API is its configuration file format rather than a
#     subset of the code itself (e.g., public subpackages, submodules, classes),
#     no change to the code itself can be considered to break backward
#     compatibility unless that change breaks the configuration file format.
#   * Implementing headline-worthy functionality (e.g., a GUI). Technically,
#     this condition breaks the Semantic Versioning schema, which stipulates
#     that *ONLY* changes breaking backward compatibility warrant major bumps.
#     But this is the real world. In the real world, significant improvements
#     are rewarded with significant version changes.
#   In either case, the minor and patch versions both reset to 0.
# * "{minor}" specifies the minor version, incremented only when implementing
#   customary functionality in a manner preserving such compatibility. In this
#   case, the patch version resets to 0.
# * "{patch}" specifies the patch version, incremented only when correcting
#   outstanding issues in a manner preserving such compatibility.
#
# When in doubt, increment only the minor version and reset the patch version.
# For further details, see http://semver.org.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

VERSION = '0.8.4'
'''
Human-readable application version as a ``.``-delimited string.
'''


CODENAME = 'Kindlier Kaufmann'
'''
Human-readable code name associated with the current version of this
application.

This code name consists of an arbitrary adjective followed by the last name of
an arbitrary academic associated with field of bioelectricity whose first letter
is the same as the first letter of that adjective.

See Also
----------
:download:`/doc/rst/RELEASE.md`
    Further details on this incredible code name schema.
'''


VERSION_PARTS = _convert_version_str_to_tuple(VERSION)
'''
Machine-readable application version as a tuple of integers.
'''

# ....................{ METADATA ~ tag                     }....................
GIT_TAG_OLDEST_BACKWARD_COMPATIBILITY = 'v0.5.2'
'''
Oldest Git tag for which this application guarantees backward compatibility.

In the context of this application, "backward compatibility" specifically means
the capacity for versions of this application newer than this Git tag to
successfully load *all*:

* Simulation configuration files loadable by this older version.
* Pickled seeds, initializations, and simulations saved by this older version.

See Also
----------
:func:`betse_test.func.sim.test_sim.test_cli_sim_backward_compatibility`
    Functional test in our test suite programmatically guaranteeing backward
    compatibility with this older version.
'''

# ....................{ METADATA ~ synopsis                }....................
# Note that a human-readable multiline description is exposed via the top-level
# "setup.py" script. This description is inefficiently culled from the contents
# of the top-level "README.rst" file and hence omitted here. (Doing so here
# would significantly increase program startup costs with little to no benefit.)
SYNOPSIS = 'BETSE, the BioElectric Tissue Simulation Engine.'
'''
Human-readable single-line synopsis of this application.

By PyPI design, this string must *not* span multiple lines or paragraphs.
'''


DESCRIPTION = (
    'The BioElectric Tissue Simulation Engine (BETSE) is a '
    'finite volume simulator for 2D computational multiphysics problems in '
    'the life sciences -- including electrodiffusion, electro-osmosis, '
    'galvanotaxis, voltage-gated ion channels, gene regulatory networks, '
    'and biochemical reaction networks.'
)
'''
Human-readable multiline description of this application.

By :mod:`argparse` design, this string may (and typically should) span both
multiple lines and paragraphs. Note that this string is *not* published to PyPI,
which accepts reStructuredText (rst) and is thus passed the contents of the
top-level :doc:`/README` file instead.
'''

# ....................{ METADATA ~ authors                 }....................
AUTHORS = 'Alexis Pietak, Cecil Curry, et al.'
'''
Human-readable list of all principal authors of this application as a
comma-delimited string.

For brevity, this string *only* lists authors explicitly assigned copyrights.
For the list of all contributors regardless of copyright assignment or
attribution, see the top-level `AUTHORS.md` file.
'''


AUTHOR_EMAIL = 'alexis.pietak@gmail.com'
'''
Email address of the principal corresponding author (i.e., the principal author
responding to public correspondence).
'''

# ....................{ METADATA ~ urls                    }....................
URL_HOMEPAGE = 'https://gitlab.com/betse/betse'
'''
URL of this application's homepage.
'''


URL_DOWNLOAD = (
    'https://gitlab.com/betse/betse/repository/archive.tar.gz?ref=v{}'.format(
        VERSION,
    )
)
'''
URL of the source tarball for the current version of this application.

This URL assumes a tag whose name is ``v{VERSION}`` where ``{VERSION}`` is the
human-readable current version of this application (e.g., ``v0.4.0``) to exist.
Typically, no such tag exists for live versions of this application -- which
have yet to be stabilized and hence tagged. Hence, this URL is typically valid
*only* for previously released (rather than live) versions of this application.
'''

# ....................{ METADATA ~ python                  }....................
PACKAGE_NAME = NAME.lower()
'''
Fully-qualified name of the top-level Python package implementing this
application.
'''


SCRIPT_BASENAME = PACKAGE_NAME
'''
Basename of the CLI-specific Python script wrapper created by :mod:`setuptools`
installation.
'''

# ....................{ METADATA ~ private                 }....................
_IS_TESTING = False
'''
``True`` only if the active Python interpreter is running a test session (e.g.,
with the ``py.test`` test harness).

This private global is subject to change and thus *not* intended to be publicly
accessed. Consider calling the public :func:`betse.util.py.pys.is_testing`
function instead.
'''



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level facilities for **pipelining** (i.e., iteratively displaying and/or
exporting) all post-simulation exports -- including plots, animations, and
spreadsheets.
'''

# ....................{ IMPORTS                            }....................
from betse.science.phase.phasecls import SimPhase
from betse.science.phase.phaseenum import SimPhaseKind
from betse.science.export.csv.csvpipe import SimPipeExportCSVs
from betse.science.visual.anim.animpipe import AnimCellsPipe
from betse.science.visual.plot.pipe.plotpipecell import PlotCellPipe
from betse.science.visual.plot.pipe.plotpipecells import PlotCellsPipe
from betse.util.io.log import logs
from betse.util.type.types import type_check

# ....................{ PIPELINES                          }....................
@type_check
def pipeline(phase: SimPhase) -> None:
    '''
    Display and/or save all exports (e.g., comma-separated value (CSV) files,
    plots, animations) enabled in a pipelined fashion for the passed simulation
    phase by the current simulation configuration.

    Parameters
    ----------------------------
    phase: SimPhase
        Current simulation phase.
    '''

    #FIXME: This is atrocious, but unfortunately still required by "plotutil"
    #utility functions transitively called below. After refactoring all such
    #plots to leverage the "PlotCellsABC" and "LayerCellsABC" APIs instead,
    #remove this shameful kludgery.

    # String constant corresponding to the current simulation phase. Valid
    # values include:
    # * `init`, for plotting initialization phase results.
    # * `sim`, for plotting simulation phase results.
    phase.p.plot_type = None
    if phase.kind is SimPhaseKind.SEED:
        phase.p.plot_type = 'seed'
    elif phase.kind is SimPhaseKind.INIT:
        phase.p.plot_type = 'init'
    elif phase.kind is SimPhaseKind.SIM:
        phase.p.plot_type = 'sim'

    # Display and/or save all exports enabled by this configuration in
    # increasing order of expected duration (i.e., from fastest to slowest),
    # trivially improving aesthetic responsiveness for end users:
    #
    # * CSV files, typically exported faster than visuals.
    # * Single-cell plots, typically exported faster than cell cluster plots.
    # * Cell cluster plots, typically exported faster than animations.
    # * Animations.
    SimPipeExportCSVs(phase).run()
    PlotCellPipe(phase).run()
    PlotCellsPipe(phase).run()
    AnimCellsPipe(phase).run()

    # Log the directory to which all results were exported.
    logs.log_info('Results exported to: %s', phase.export_dirname)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **kernel** (i.e., central core software of the current operating
system) facilities.

Caveats
----------
**Operating system-specific logic is poor form.** Do so _only_ where necessary.
'''

# ....................{ IMPORTS                            }....................
import platform

from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.mapping.mapcls import OrderedArgsDict


# ....................{ GETTERS                            }....................
@func_cached
def get_name() -> str:
    '''
    Machine-readable name of the current kernel.

    While preserving human-readability is helpful, machine-readability is the
    principal use case for this function's return value (e.g., for use in
    dictionaries keyed on kernel type). This function returns:

    * Under Linux, `Linux`.
    * Under macOS, `Darwin`.
    * Under Windows (both Cygwin and vanilla), `Windows`.
    * Under all other platforms, the string returned by the `platform.system()`
      function.
    '''

    # Return platform.system() as is, which appears to exactly corresponding to
    # the name of the current kernel on all platforms (e.g., "Darwin", "Linux").
    return platform.system()


@func_cached
def get_version() -> str:
    '''
    Human-readable `.`-delimited version specifier of the current kernel.

    This function returns:

    * Under Linux, the current Linux kernel version (e.g., `4.1.15-gentoo-r1`),
    * Under macOS, the current XNU kernel version (e.g., `13.0.0`).
    * Under Windows, the current Windows API version (e.g., `6.2.9200`).
    * Under all other platforms, the string returned by both the
      `get_version()` and `platform.release()` functions.
    '''

    # Avoid circular import dependencies.
    from betse.util.os import oses

    # Version specifier to be returned, defaulting to that returned by
    # platform.release(). While this typically corresponds to the low-level
    # version of the current kernel (e.g., "4.1.15" under Linux, "13.0.0" under
    # macOS), this is *NOT* necessarily the case (e.g., "8" rather than
    # "6.2.9200" under Windows). Hence, this is only a fallback.
    kernel_version = platform.release()

    # If the Windows-specific platform.win32_ver() function is available, return
    # the second element of the 4-tuple "(release, version, csd, ptype)"
    # returned by this function (e.g., "6.2.9200"). Since platform.release()
    # returns the high-level operating system version (e.g., "8"), this version
    # is ignored when feasible.
    if oses.is_windows() and hasattr(platform, 'win32_ver'):
        kernel_version = platform.win32_ver()[1]
    # If Linux, macOS, or other platforms, accept the default version specifier
    # returned by platform.release() (e.g., "4.1.15", "13.0.0").

    # Return this version.
    return kernel_version

# ....................{ GETTERS ~ metadata                 }....................
def get_metadata() -> OrderedArgsDict:
    '''
    Ordered dictionary synopsizing the current kernel.
    '''

    # Return this dictionary.
    return OrderedArgsDict(
        'name', get_name(),
        'version', get_version(),
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Fixture classes encapsulating test-related simulation configurations.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta
from betse.util.type.decorator.deccls import abstractproperty
from betse.util.type.types import type_check
from py._path.local import LocalPath

# ....................{ SUPERCLASSES                       }....................
class SimConfTestABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all simulation configuration context subclasses, each
    encapsulating state and metadata for a test-related simulation
    configuration.

    Simulation configuration fixtures typically return instances of this class
    as a means of communicating this context to other fixtures and tests.

    Attributes
    ----------
    conf_dirname : str
        Absolute pathname of the parent directory containing the simulation
        configuration file encapsulated by this context.
    conf_filename : str
        Absolute pathname of the simulation configuration file encapsulated by
        this context.

    See Also
    ----------
    https://py.readthedocs.org/en/latest/path.html
        Official :mod:`py.path` submodule documentation.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, conf_filename: str) -> None:
        '''
        Initialize this simulation configuration context.

        Parameters
        ----------
        conf_filename : str
            Absolute pathname of the simulation configuration file encapsulated
            by this context. Although this file need *not* physically exist
            before this object is instantiated (i.e., before this method is
            called), the subclass should ensure this file exists once this
            object is instantiated (i.e., immediately before the subclass
            implementation of this method returns).
        '''

        # Defer heavyweight imports.
        from betse.util.path import pathnames

        # Classify all passed parameters publicly, permitting access by tests.
        self.conf_filename = conf_filename
        self.conf_dirname = pathnames.get_dirname(self.conf_filename)

    # ..................{ CONTEXTS                           }..................
    def context(self) -> 'contextlib.contextmanager':
        '''
        Context manager changing the current working directory (CWD) of the
        current test to the directory containing this configuration file for the
        duration of this context.

        Default simulation configuration paths are relative to the directory
        containing the simulation configuration file: namely, this temporary
        directory. Changing directories resolves these paths to this directory.
        (Failing to do so would incorrectly resolve these paths to the current
        directory, with predictably disastrous outcomes.) While this class
        could instead globally search-and-replace all relative simulation
        configuration paths with absolute paths, doing so would be considerably
        more complex, fragile, and error-prone than simply changing directories.
        '''

        # Defer heavyweight imports.
        from betse.util.os.shell import shelldir

        # Defer to the generator returned by the following utility function.
        return shelldir.setting_cwd(self.conf_dirname)

    # ..................{ SUBCLASS                           }..................
    # Subclasses are required to define the following read-only properties.

    @abstractproperty
    def p(self) -> 'betse.science.parameters.Parameters':
        '''
        High-level simulation configuration encapsulated by this test wrapper.
        '''

        pass

# ....................{ SUBCLASSES                         }....................
#FIXME: Most use of the increasingly obsolete "SimConfTestInternal.config"
#wrapper attribute (both here and everywhere else) should be replaced by use of
#the new "SimConfTestInternal.p" property, which increasingly provides all test
#functionality.
class SimConfTestInternal(SimConfTestABC):
    '''
    Simulation configuration context subclass encapsulating a temporary
    simulation configuration file internally isolated to the current test.

    Since this configuration is guaranteed to be test-specific and hence safely
    modifiable, caller fixtures and tests are advised to modify the contents of
    this configuration (e.g., to exercise specific feature sets and edge cases).

    Attributes
    ----------
    config : SimConfigWrapper
        Simulation configuration wrapper wrapping the low-level dictionary
        deserialized from the YAML-formatted simulation configuration file with
        path :attr:`conf_filepath`. Note the contents of this dictionary may
        be desynchronized from those of this file. For efficiency, callers may
        modify this dictionary to suite test requirements *before* reserializing
        this dictionary back to this file.

    Attributes (Path)
    ----------
    conf_filepath : LocalPath
        Absolute path of a temporary simulation configuration file specific to
        the parent fixture as a :class:`py.path.local` instance, defining an
        object-oriented superset of the non-object-oriented :mod:`os.path`
        module.

    See Also
    ----------
    https://py.readthedocs.org/en/latest/path.html
        Official :mod:`py.path` submodule documentation.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, conf_filepath: LocalPath) -> None:
        '''
        Initialize this simulation configuration context.

        This method (in order):

        #. Copies BETSE's default simulation configuration file, complete with
           all external assets (e.g., geometry masks) referenced and required by
           this file, to the passed path.
        #. Sanitizes the copied simulation configuration file for all child
           fixtures and tests by unconditionally disabling options either
           requiring interactive input *or* displaying interactive output.

        This method does *not* minimize the space and time costs of running
        simulations configured by this configuration. Doing so is typically
        desirable but can obscure edge-case issues, including computational
        instability produced by the default non-minified time steps.

        Caveats
        ----------
        For efficiency, note that the configuration changes applied by this
        method (listed above) reside *only* in-memory; they have yet to be
        written back to disk. Callers are required to do so manually (e.g., the
        ``is_overwriting_config`` parameter passed to the
        :meth:`CLISimTester.run_subcommands` method).

        Parameters
        ----------
        conf_filepath : LocalPath
            Absolute path to which this method will copy the default simulation
            configuration file as a :class:`py.path.local` instance. If this
            file already exists, an exception is raised.
        '''

        # Defer heavyweight imports. This subclass inherits a class defined by
        # the main codebase and is hence *NOT* safely importable above.
        from betse_test.fixture.simconf.simconfwrapper import (
            SimConfigTestWrapper)

        # Initialize our superclass with the absolute pathname of this path.
        super().__init__(conf_filename=str(conf_filepath))

        # Classify the passed parameters. While the "self.config" object
        # classified below provides this filename as a low-level string, this
        # high-level "py.path.local" instance is useful in fixtures and tests.
        self.conf_filepath = conf_filepath

        # Configuration deserialized from this file, reducing this filename from
        # a high-level "py.path.local" instance to a low-level string.
        self.config = SimConfigTestWrapper.wrap_new_default(
            filename=self.conf_filename)

        # Sanitize this configuration for all child fixtures and tests.
        self.config.disable_interaction()

    # ..................{ SUPERCLASS                         }..................
    @property
    def p(self) -> 'betse.science.parameters.Parameters':

        return self.config.p


class SimConfTestExternal(SimConfTestABC):
    '''
    Simulation configuration context subclass encapsulating an external
    simulation configuration file residing at any directory and hence *not*
    necessarily isolated to the current test.

    Attributes
    ----------
    _p : Parameters
        High-level simulation configuration encapsulating this external
        simulation configuration file.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this simulation configuration context.
        '''

        # Defer heavyweight imports.
        from betse.science.parameters import Parameters

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # In-memory simulation configuration deserialized from this file.
        self._p = Parameters().load(self.conf_filename)

    # ..................{ SUPERCLASS                         }..................
    @property
    def p(self) -> 'betse.science.parameters.Parameters':

        return self._p



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

"""
Controls a gene regulatory network.

Creates and electrodiffuses a suite of customizable general gene products in the
BETSE ecosystem, where the gene products are assumed to activate and/or inhibit
the expression of other genes (and therefore the production of other gene
products) in the gene regulatory network (GRN).
"""

import numpy as np
from betse.science import filehandling as fh
from betse.util.io.log import logs
from betse.util.path import pathnames
from betse.science.chemistry.networks import MasterOfNetworks
from betse.science.config import confio
from betse.science.chemistry.netplot import set_net_opts
# from betse.science import sim_toolbox as stb
from betse.science.visual.plot import plotutil as viz
import matplotlib.pyplot as plt
from betse.science.organelles.microtubules import Mtubes


class MasterOfGenes(object):

    def __init__(self, p):

        pass


    def read_gene_config(self, sim, cells, p):

        # Previously loaded GRN-specific configuration file as a dictionary.
        config_dic = p.grn.conf

        # obtain specific sub-dictionaries from the config file:
        substances_config = config_dic['biomolecules']
        reactions_config = config_dic.get('reactions', None)
        transporters_config = config_dic.get('transporters', None)
        channels_config = config_dic.get('channels', None)
        modulators_config = config_dic.get('modulators', None)

        # initialize the substances of metabolism in a core field encapsulating
        # Master of Molecules:
        self.core = MasterOfNetworks(sim, cells, substances_config, p)

        # Time dilation:
        self.core.time_dila = float(config_dic.get('time dilation factor', 1.0))

        # reset microtubules?
        self.reset_MT = config_dic.get('reset microtubules', False)

        # recalculate fluid flow?
        self.recalc_fluid = config_dic.get('recalculate fluid', False)

        # read in substance properties from the config file, and initialize basic properties:
        self.core.read_substances(sim, cells, substances_config, p)
        self.core.tissue_init(sim, cells, substances_config, p)


        if reactions_config is not None:
            # initialize the reactions of metabolism:
            self.core.read_reactions(reactions_config, sim, cells, p)
            self.core.write_reactions()
            self.core.create_reaction_matrix()
            self.core.write_reactions_env()
            self.core.create_reaction_matrix_env()

            self.reactions = True

        else:
            self.core.create_reaction_matrix()
            self.core.create_reaction_matrix_env()
            self.reactions = False

        # initialize transporters, if defined:
        if transporters_config is not None:
            self.core.read_transporters(transporters_config, sim, cells, p)
            self.core.write_transporters(sim, cells, p)

            self.transporters = True

        else:
            self.transporters = False

        # initialize channels, if desired:
        if channels_config is not None:
            self.core.read_channels(channels_config, sim, cells, p)
            self.channels = True

        else:
            self.channels = False

        # initialize modulators, if desired:

        if modulators_config is not None:
            self.core.read_modulators(modulators_config, sim, cells, p)
            self.modulators = True

        else:
            self.modulators = False

        # read in network plotting options:
        self.core.net_plot_opts = config_dic.get('network plotting', None)

        # set plotting options for the network:
        set_net_opts(self.core, self.core.net_plot_opts, p)

        # after primary initialization, check and see if optimization required:
            # after primary initialization, check and see if optimization required:
        opti = config_dic['optimization']['optimize network']
        self.core.opti_N = config_dic['optimization']['optimization steps']
        self.core.opti_method = config_dic['optimization']['optimization method']
        self.core.target_vmem = float(config_dic['optimization']['target Vmem'])
        self.core.opti_T = float(config_dic['optimization']['optimization T'])
        self.core.opti_step = float(config_dic['optimization']['optimization step'])
        # self.core.opti_run = config_dic['optimization']['run from optimization']

        if opti is True:
            logs.log_info("The Gene Network is being analyzed for optimal rates...")
            self.core.optimizer(sim, cells, p)
            self.reinitialize(sim, cells, p)

    def reinitialize(self, sim, cells, p):

        # Previously loaded GRN-specific configuration file as a dictionary.
        config_dic = p.grn.conf

        # Time dilation:
        self.core.time_dila = float(config_dic.get('time dilation factor', 1.0))

        # reset microtubules?
        self.reset_MT = config_dic.get('reset microtubules', False)

        # recalculate fluid flow?
        self.recalc_fluid = config_dic.get('recalculate fluid flow', False)

        # obtain specific sub-dictionaries from the config file:
        substances_config = config_dic['biomolecules']
        reactions_config = config_dic.get('reactions', None)
        transporters_config = config_dic.get('transporters', None)
        channels_config = config_dic.get('channels', None)
        modulators_config = config_dic.get('modulators', None)

        self.core.tissue_init(sim, cells, substances_config, p)

        if reactions_config is not None:
            # initialize the reactions of metabolism:
            self.core.read_reactions(reactions_config, sim, cells, p)
            self.core.write_reactions()
            self.core.create_reaction_matrix()
            self.core.write_reactions_env()
            self.core.create_reaction_matrix_env()

            self.reactions = True

        else:
            self.core.create_reaction_matrix()
            self.core.create_reaction_matrix_env()
            self.reactions = False

        # initialize transporters, if defined:
        if transporters_config is not None:
            self.core.read_transporters(transporters_config, sim, cells, p)
            self.core.write_transporters(sim, cells, p)

            self.transporters = True

        else:
            self.transporters = False

        # initialize channels, if desired:
        if channels_config is not None:
            self.core.read_channels(channels_config, sim, cells, p)
            self.channels = True

        else:
            self.channels = False

        # initialize modulators, if desired:
        if modulators_config is not None:
            self.core.read_modulators(modulators_config, sim, cells, p)
            self.modulators = True

        else:
            self.modulators = False

    def run_core_sim(self, sim, cells, p):

        # set molecules to not affect charge for sim-grn test-drives:
        p.substances_affect_charge = False

        # Set sim-grn specific time step and sampling:
        p.dt = p.grn_dt
        p.init_time_total = p.grn_total_time
        p.init_time_sampling = p.grn_tsample

        p.init_tsteps = int(p.init_time_total / p.dt)
        p.resample = p.init_time_sampling
        p.total_time = p.init_time_total

        # Number of time steps (including sampled and unsampled) between each
        # unsampled time step, including that unsampled time step itself.
        p.t_resample = p.resample / p.dt

        # specify a time vector
        loop_time_step_max = p.init_tsteps
        # Maximum number of seconds simulated by the current run.
        loop_seconds_max = loop_time_step_max * p.dt
        # Time-steps vector appropriate for the current run.
        tt = np.linspace(0, loop_seconds_max, loop_time_step_max)

        tsamples = set()
        i = 0
        while i < len(tt) - p.t_resample:
            i = int(i + p.t_resample)
            # logs.log_debug('Time sample i: {!r}'.format(i))
            tsamples.add(tt[i])

        # if p.grn_runmodesim:
        self.reinitialize(sim, cells, p)

        self.core.clear_cache()
        self.time = []

        self.mod_after_cut = False # set this to false

        if self.recalc_fluid:  # If user requests the GRN recalculate/calculate fluid:

            logs.log_info("Calculating fluid in terms of endogenous currents...")

            p.fluid_flow = True  # turn fluid flow on (in case it was off)

            # calculate or re-calculate fluid flow in terms of curl-component of current:
            cc = sim.cc_env.mean(axis=0).reshape(cells.X.shape)
            zz = sim.zs.mean()

            sim.u_env_x = -sim.J_env_x / (p.F * cc * zz)
            sim.u_env_y = -sim.J_env_y / (p.F * cc * zz)

        if p.use_microtubules:

            sim.mtubes.reinit(cells, p)

            self.mtubes_x_time = []
            self.mtubes_y_time = []

            if self.reset_MT:
                logs.log_info("Resetting microtubules for sim-grn simulation...")
                sim.mtubes = Mtubes(sim, cells, p)

        for t in tt:

            if self.transporters:
                self.core.run_loop_transporters(t, sim, cells, p)

            self.core.run_loop(t, sim, cells, p)


            if p.use_microtubules: # update the microtubules:
                sim.mtubes.update_mtubes(cells, sim, p)

            if p.grn_runmodesim is True and t > p.cut_time and self.mod_after_cut is False:

                sim.dyna.runAllDynamics(sim, cells, p, t)

                if sim.dyna.event_cut.is_fired and self.mod_after_cut is False: # if a cutting event has just been run:

                    self.core.mod_after_cut_event(sim.target_inds_cell_o, sim.target_inds_mem_o, sim, cells, p)
                    logs.log_info(
                        "Redefining dynamic dictionaries to point to the new sim...")
                    self.core.redefine_dynamic_dics(sim, cells, p)

                    logs.log_info(
                        "Reinitializing the gene regulatory network for simulation...")
                    self.reinitialize(sim, cells, p)
                    # self.core.clear_cache()
                    sim.uxmt, self.uymt = sim.mtubes.mtubes_to_cell(cells, p)

                    self.mod_after_cut = True  # set the boolean to avoid repeat action


            if t in tsamples:
                sim.time.append(t)

                logs.log_info('------------------' + str(np.round(t,3)) +' s --------------------')
                self.time.append(t)
                self.core.write_data(sim, cells, p)
                self.core.report(sim, p)

                if p.use_microtubules:
                    # microtubules:
                    self.mtubes_x_time.append(sim.mtubes.mtubes_x * 1)
                    self.mtubes_y_time.append(sim.mtubes.mtubes_y * 1)



        logs.log_info('Saving simulation...')
        datadump = [self, cells, p]
        fh.saveSim(p.grn_pickle_filename, datadump)
        self.core.init_saving(cells, p, plot_type='grn', nested_folder_name='RESULTS')

        # microtubules plot------------------------------------------------------------------------
        if p.use_microtubules:

            logs.log_info("Plotting microtubules used in GRN simulation...")

            plt.figure()
            ax = plt.subplot(111)

            umtx, umty = sim.mtubes.mtubes_to_cell(cells, p)

            plt.figure()
            ax = plt.subplot(111)

            viz.plotVectField(
                umtx,
                umty,
                cells,
                p,
                plot_ecm=False,
                title='Final Microtubule Alignment Field',
                cb_title='Aligned MT Fraction',
                colorAutoscale=False,
                minColor=0.0,
                maxColor=1.0,
            )

            # viz.mem_quiver(
            #     sim.mtubes.mtubes_x,
            #     sim.mtubes.mtubes_y,
            #     ax,
            #     cells,
            #     p,
            # )

            ax.set_xlabel('X-Distance [um]')
            ax.set_ylabel('Y-Distance [um]')
            ax.set_title('Microtubule arrangement in cells')

            if p.autosave is True:
                savename = self.core.imagePath + 'Microtubules' + '.png'
                plt.savefig(savename, format='png', transparent=True)

            if p.turn_all_plots_off is False:
                plt.show(block=False)

        if self.recalc_fluid:

            logs.log_info("Plotting fluid flow used in GRN simulation")

            plt.figure()
            ax = plt.subplot(111)

            viz.plotStreamField(
                1e9 * sim.u_env_x,
                1e9 * sim.u_env_y,
                cells, p,
                plot_ecm=True,
                title='Final Fluid Velocity in Environment',
                cb_title='Velocity [nm/s]'
            )

            if p.autosave is True:
                savename = self.core.imagePath + 'Fluid_ECM' + '.png'
                plt.savefig(savename, format='png', transparent=True)

            if p.turn_all_plots_off is False:
                plt.show(block=False)


        self.core.export_eval_strings(p)
        self.core.export_equations(p)
        message = 'Gene regulatory network simulation saved to' + ' ' + p.grn_pickle_filename
        logs.log_info(message)

        logs.log_info('-------------------Simulation Complete!-----------------------')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level object facilities.
'''

# ....................{ IMPORTS                            }....................
import inspect, platform
from betse.exceptions import (
    BetseAttributeException, BetseMethodException, BetseTypeException)
from betse.util.type import types
from betse.util.type.types import (
    type_check, CallableTypes, CallableOrNoneTypes, ClassType, GeneratorType)

# ....................{ EXCEPTIONS                         }....................
@type_check
def die_unless_instance(obj: object, cls: ClassType) -> None:
    '''
    Raise an exception unless the passed object is an instance of the passed
    class.

    Parameters
    ----------
    obj : object
        Object to be validated.
    cls : ClassType
        Class to be validated.
    '''

    if not isinstance(obj, cls):
        raise BetseTypeException(
            'Object {!r} not an instance of class {!r}'.format(obj, cls))

# ....................{ EXCEPTIONS ~ attr                  }....................
@type_check
def die_unless_class(obj: object, *class_names: str) -> None:
    '''
    Raise an exception unless the passed object provides all classes with the
    passed names.

    Parameters
    ----------
    obj : object
        Object to test for these classes.
    class_names : tuple[str]
        Tuple of the names of all classes to test this object for.

    Raises
    ----------
    BetseMethodException
        If one or more such classes are *not* bound to this object.
    '''

    for class_name in class_names:
        if not is_class(obj=obj, class_name=class_name):
            raise BetseMethodException(
                'Object "{}" class "{}" undefined.'.format(obj, class_name))


@type_check
def die_unless_method(obj: object, *method_names: str) -> None:
    '''
    Raise an exception unless the passed object provides all methods with the
    passed names.

    Parameters
    ----------
    obj : object
        Object to test for these methods.
    method_names : tuple[str]
        Tuple of the names of all methods to test this object for.

    Raises
    ----------
    BetseMethodException
        If one or more such methods are *not* bound to this object.
    '''

    for method_name in method_names:
        if not is_method(obj=obj, method_name=method_name):
            raise BetseMethodException(
                'Object "{}" method {}() undefined.'.format(obj, method_name))

# ....................{ TESTERS                            }....................
@type_check
def is_attr(obj: object, attr_name: str) -> bool:
    '''
    ``True`` only if the passed object provides an attribute of arbitrary type
    with the passed name.

    This function is a convenient synonym of the :func:`hasattr` builtin,
    provided entirely as a caller convenience.

    Parameters
    ----------
    obj : object
        Object to be tested for this attribute.
    attr_name : str
        Name of the attribute to be tested for.

    Returns
    ----------
    bool
        ``True`` only if this attribute is bound to this object.
    '''

    return hasattr(obj, attr_name)


@type_check
def is_class(obj: object, class_name: str) -> bool:
    '''
    ``True`` only if the passed object provides a class with the passed name.

    Parameters
    ----------
    obj : object
        Object to test for this class.
    class_name : str
        Name of the class to test this object for.

    Returns
    ----------
    bool
        `True` only if a class with this name is bound to this object.
    '''

    # Attribute with this name in this object if any or "None" otherwise.
    cls = getattr(obj, class_name, None)

    # Return whether this attribute is a class. Since "None" is guaranteed to
    # *NOT* be a class, this simple test produces the expected result in all
    # possible cases.
    return types.is_class(cls)


@type_check
def is_method(obj: object, method_name: str) -> bool:
    '''
    ``True`` only if the passed object provides a method with the passed name.

    Parameters
    ----------
    obj : object
        Object to test for this method.
    method_name : str
        Name of the method to test this object for.

    Returns
    ----------
    bool
        ``True`` only if a method with this name is bound to this object.
    '''

    # Method with this name bound to this object if any or "None" otherwise.
    method = get_method_or_none(obj=obj, method_name=method_name)

    # Return true only if this method exists.
    return method is not None

# ....................{ TESTERS ~ pure                     }....................
# Tester distinguishing pure-Python from C-based class instances. Doing so is
# surprisingly non-trivial and, indeed, technically feasible with 100% accuracy
# *ONLY* under the official CPython interpreter. Why? Because the most accurate
# means of portably implementing this test in a cross-interpreter manner is to
# test whether the passed class instance defines the reserved "__dict__" or
# "__slots__" attributes. Whereas all pure-Python class instances are guaranteed
# to declare one or the other, most C-based class instances define neither. Of
# course, "most" is hardly "all." A C-based class may optionally request a
# "__dict__" attribute via "tp_dictoffset", resulting in false negatives in
# uncommon edge cases.
#
# If *NOT* running under the official CPython interpreter, this is the strongest
# possible implementation of this test; if running under the official CPython
# interpreter, however, this implementation may be strengthened by detecting
# whether or not the class of the passed class instance is a "heap type" (as
# defined by the "_Py_TPFLAGS_HEAPTYPE" docstring below). Only pure-Python
# classes are heap types, preserving a one-to-one correspondence between tester
# results and underlying interpreter reality.
#
# If the active Python interpreter is the official CPython implementation,
# prefer a more reliable CPython-specific solution guaranteed to succeed. Note
# that, to avoid circular import dependencies, this conditional unavoidably
# duplicates the existing betse.util.py.interpreters.is_cpython() tester.
if platform.python_implementation() == 'CPython':
    _Py_TPFLAGS_HEAPTYPE = (1<<9)
    '''
    Magic number defined by the `Include/object.h` C header in the official
    CPython codebase.

    CPython ORs the `__flags__` bit field of the class object for each **heap
    type** (i.e., pure-Python class whose type structure is dynamically
    allocated at interpreter runtime rather than a C-based class whose type
    structure is statically preallocated at either interpreter or C extension
    compile time).
    '''

    # To reduce duplication, this tester's docstring is dynamically set below.
    @type_check
    def is_pure_python(obj: object) -> bool:

        # If the passed object is *NOT* a class, this object *MUST* be a class
        # instance. (By Python 3.x design, all objects -- including primitive
        # builtins such as "int" and "bool" -- are one or the other). In this
        # case, obtain the class this instance is an instance of.
        cls = obj if isinstance(obj, type) else type(obj)

        # Return True only if this class is a heap type, as indicated by its
        # heap type bit flag being non-zero.
        return bool(cls.__flags__ & _Py_TPFLAGS_HEAPTYPE)
# Else, fallback to a CPython-agnostic solution typically but *NOT* necessarily
# succeeding. For all real-world objects of interest, this is effectively
# successful. Edge cases exist but are suitably rare. See the commentary above.
else:
    # To reduce duplication, this tester's docstring is dynamically set below.
    @type_check
    def is_pure_python(obj: object) -> bool:

        # If the passed object is a class, return True only if this class
        # defines either the "__dict__" or "__slots__" attributes.
        # Unfortunately, the trivial test "hasattr(cls, '__dict__')" does *NOT*
        # suffice to decide whether this class defines the "__dict__" attribute.
        # Why? Because this attribute unconditionally exists for *ALL* classes,
        # pure-Python and C-based alike. Hence:
        #
        #     >>> hasattr(int, '__dict__')
        #     True
        #
        # For unknown and presumably banal reasons, the dir() builtin strips the
        # "__dict__" attribute name from its returned list only for C-based
        # classes. Hence, detecting whether a class is pure-Python or C-based in
        # a cross-interpreter manner reduces to iteratively searching the list
        # returned by dir() for this name.
        #
        # This constraint does *NOT* extend to the "__slots__" attribute, which
        # exists if and only if this class explicitly defines this attribute.
        # Are we having fun yet?
        if isinstance(obj, type):
            return '__dict__' in dir(obj) or hasattr(obj, '__slots__')
        # Else, this object *MUST* be a class instance. In this case, return
        # True only if similar conditions hold. Since a class instance defines
        # the "__dict__" attribute only if this instance is a pure-Python
        # instance whose class does *NOT* define the "__slots__" attribute, this
        # conditional may efficiently lookup the "__dict__" attribute directly
        # rather than inefficiently defer to the dir() builtin. It is fun!
        else:
            return hasattr(obj, '__dict__') or hasattr(obj, '__slots__')

# Docstring dynamically set for the tester defined above.
is_pure_python.__doc__ = '''
``True`` if the passed object is either a pure-Python class or instance of such
a class *or* ``False`` otherwise (i.e., if this object is either a C-based class
or instance of such a class, either builtin or defined by a C extension,).

Parameters
----------
obj : object
    Object to be tested.

Returns
----------
bool
    ``True`` only if this object is a pure-Python class _or_ instance of such a
    class.

See Also
----------
https://stackoverflow.com/a/41012823/2809027
    Stackoverflow answer strongly inspiring this implementation.
'''


def is_c_based(obj: object) -> bool:
    '''
    ``True`` if the passed object is either a C-based class or instance of such a
    class (either builtin or defined by a C extension) *or* ``False`` otherwise
    (i.e., if this object is either a pure-Python class or instance of such a
    class).

    Parameters
    ----------
    obj : object
        Object to be tested.

    Returns
    ----------
    bool
        ``True`` only if this object is a C-based class _or_ instance of such a
        class.
    '''

    return not is_pure_python(obj)

# ....................{ GETTERS ~ metadata                 }....................
def get_class_name(obj: object) -> str:
    '''
    Unqualified name of either the passed object if this object is itself a
    class *or* the class of this object otherwise (i.e., if this object is *not*
    a class).

    Parameters
    ----------
    obj : object
        Object to retrieve this class name for.

    Returns
    ----------
    str
        Unqualified name of this class.
    '''

    # This class if this object is a class *OR* this object's class otherwise.
    cls = obj if isinstance(obj, ClassType) else type(obj)

    # Else, return this class' name.
    return cls.__name__


def get_class_module_name(obj: object) -> str:
    '''
    Fully-qualified name of the module defining either the passed object if this
    object is itself a class *or* the class of this object otherwise (i.e., if
    this object is *not* a class).

    Parameters
    ----------
    obj : object
        Object to retrieve this module name for.

    Returns
    ----------
    str
        Fully-qualified name of this module.

    Raises
    ----------
    BetseTypeException
        If this class has no ``__module__`` attribute, which should ideally
        *never* happen.
    '''

    # This class if this object is a class *OR* this object's class otherwise.
    cls = obj if isinstance(obj, ClassType) else obj.__class__

    # If this module does *NOT* provide the special "__file__" attribute, raise
    # an exception. All modules including builtin modules should provide this.
    if not hasattr(cls, '__module__'):
        raise BetseTypeException(
            'Class "{}.__module__" attribute not found.'.format(
                cls.__name__))

    # Else, return this attribute's value.
    return cls.__module__

# ....................{ GETTERS : attr                     }....................
@type_check
def get_attr(obj: object, attr_name: str) -> object:
    '''
    Attribute with the passed name bound to the passed object if any *or* raise
    an exception otherwise.

    Parameters
    ----------
    obj : object
        Object to obtain this attribute from.
    attr_name : str
        Name of the attribute to be obtained.

    Returns
    ----------
    object
        Attribute with this name bound to this object.

    Raises
    ----------
    BetseAttributeException
        If no such attribute is bound to this object.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Attribute with this name in this object if any or the sentinel otherwise.
    attr = get_attr_or_sentinel(obj, attr_name)

    # If no such attribute exists, raise an exception.
    if attr is SENTINEL:
        raise BetseAttributeException(
            'Object attribute "{}.{}" undefined.'.format(
                obj.__class__.__name__, attr_name))

    # Else, this attribute exists. Return this attribute as is.
    return attr


@type_check
def get_attr_or_sentinel(obj: object, attr_name: str) -> object:
    '''
    Value of the attribute with the passed name bound to the passed object if
    any *or* :attr:`betse.util.type.obj.sentinels.SENTINEL` otherwise,
    permitting callers to distinguish between attributes that do *not* exist and
    attributes that do exist but whose values are ``None``.

    Parameters
    ----------
    obj : object
        Object to obtain this attr from.
    attr_name : str
        Name of the attribute to be obtained.

    Returns
    ----------
    object
        Value of this attribute if any *or* the sentinel singleton otherwise.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Return this attribute if any or the sentinel otherwise.
    return getattr(obj, attr_name, SENTINEL)

# ....................{ GETTERS : method                   }....................
@type_check
def get_method(obj: object, method_name: str) -> CallableTypes:
    '''
    Method with the passed name bound to the passed object if any *or* raise
    an exception otherwise.

    Parameters
    ----------
    obj : object
        Object to obtain this method from.
    method_name : str
        Name of the method to be obtained.

    Returns
    ----------
    CallableTypes
        Method with this name bound to this object.

    Raises
    ----------
    BetseMethodException
        If no such method is bound to this object.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Attribute with this name in this object if any or the sentinel otherwise.
    # To raise human-readable exceptions, this attribute is *NOT* retrieved via
    # the higher-level get_method_or_none() method; doing so would obscure
    # whether this attribute does not exist or does but is not a method.
    method = get_attr_or_sentinel(obj, method_name)

    # If no such attribute exists, raise an exception.
    if method is SENTINEL:
        raise BetseMethodException('Method {}.{}() undefined.'.format(
            obj.__class__.__name__, method_name))

    # If this attribute exists but is *NOT* a method, raise an exception.
    if not callable(method):
        raise BetseMethodException(
            'Object attribute "{}.{}" not a method: {!r}'.format(
            obj.__class__.__name__, method_name, method))

    # Else, this attribute is a method. Return this method as is.
    return method


@type_check
def get_method_or_none(obj: object, method_name: str) -> CallableOrNoneTypes:
    '''
    Method with the passed name bound to the passed object if any *or* ``None``
    otherwise.

    Parameters
    ----------
    obj : object
        Object to obtain this method from.
    method_name : str
        Name of the method to be obtained.

    Returns
    ----------
    CallableOrNoneTypes
        This method if any *or* ``None`` otherwise.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Attribute with this name in this object if any or the sentinel otherwise.
    method = get_attr_or_sentinel(obj, method_name)

    # If this attribute is a method, return this attribute; else, return "None".
    return method if method is not SENTINEL and callable(method) else None

# ....................{ ITERATORS ~ attrs                  }....................
@type_check
def iter_attrs_matching(
    obj: object, predicate: CallableTypes) -> GeneratorType:
    '''
    Generator yielding 2-tuples of the name and value of each attribute bound to
    the passed object whose name and/or value matches the passed predicate (in
    ascending lexicographic order of attribute name).

    Parameters
    ----------
    obj : object
        Object to yield all matching attribute of.
    predicate : CallableTypes
        Callable iteratively passed both the name and value of each attribute
        bound to this object, returning ``True`` only if that name and/or value
        matches this predicate.

    Yields
    ----------
    (str, object)
        2-tuple ``(attr_name, attr_value)`` of the name and value of each
        matching attribute bound to this object (in ascending lexicographic
        order of attribute name).
    '''

    # Return a generator comprehension...
    return (
        # Yielding this attribute's name and value...
        (attr_name, attr_value)
        # For the name and definition of each attribute bound to this object...
        for attr_name, attr_value in inspect.getmembers(obj)
        # If this attribute matches this predicate.
        if predicate(attr_name, attr_value)
    )

# ....................{ ITERATORS ~ methods                }....................
#FIXME: Reimplement in terms of iter_attrs_matching().
def iter_methods(obj: object) -> GeneratorType:
    '''
    Generator yielding a 2-tuple of the name and value of each method bound to
    the passed object (in ascending lexicographic order of method name).

    This includes:

    * All methods statically registered in this object's internal dictionary
      (e.g., ``__dict__`` in unslotted objects), including both:
      * Builtin methods, whose names are both prefixed and suffixed by ``__``.
      * Custom methods, whose names are *not* prefixed and suffixed by ``__``.
    * All methods dynamically returned by **property methods** (i.e., methods
      decorated by the :func:`property` decorator).

    This excludes:

    * All property methods themselves.
    * All methods dynamically defined by this object's ``__getattr__()`` method
      or related runtime magic, which cannot (by definition) be inspected by
      this generator.

    Parameters
    ----------
    obj : object
        Object to iterate all methods of.

    Yields
    ----------
    (method_name, method_value)
        2-tuple of the name and value of each method bound to this object (in
        ascending lexicographic order of method name).

    Raises
    ----------
    Exception
        If any method of this object is a property whose
        :func:`property`-decorated method raises an exception. To avoid handling
        such exceptions, consider instead calling the
        :func:`iter_methods_custom_simple` generator excluding all properties.
    '''

    # Return a generator comprehension...
    return (
        # Yielding this method's name and definition...
        (attr_name, attr_value)
        # For the name and value of each attribute of this object...
        for attr_name, attr_value in inspect.getmembers(obj)
        # If this attribute is a method.
        if callable(attr_value)
    )


#FIXME: Reimplement in terms of iter_attrs_matching().
#FIXME: Generalize the passed predicate to accept two rather than one
#parameters, as with the predicate accepted by iter_attrs_matching().
@type_check
def iter_methods_matching(
    obj: object, predicate: CallableTypes) -> GeneratorType:
    '''
    Generator yielding 2-tuples of the name and value of each method bound to
    the passed object whose method name matches the passed predicate (in
    ascending lexicographic order of method name).

    Parameters
    ----------
    obj : object
        Object to yield all matching methods of.
    predicate : CallableTypes
        Callable iteratively passed the name of each method bound to this
        object, returning ``True`` only if that name matches this predicate.

    Yields
    ----------
    (method_name, method_value)
        2-tuple of the name and value of each matching method bound to this
        object (in ascending lexicographic order of method name).

    See Also
    ----------
    :func:`iter_methods`
        Further details.
    '''

    # Return a generator comprehension...
    return (
        # Yielding this method's name and definition...
        (method_name, method)
        # For the name and definition of each method bound to this object...
        for method_name, method in iter_methods(obj)
        # If this method matches this predicate.
        if predicate(method_name)
    )


def iter_methods_custom(obj: object) -> GeneratorType:
    '''
    Generator yielding 2-tuples of the name and value of each **non-builtin
    method** (i.e., method whose name is *not* both prefixed and suffixed by
    `__`) bound to the passed object (in ascending lexicographic order of
    method name).

    Parameters
    ----------
    obj : object
        Object to yield all non-builtin methods of.

    Yields
    ----------
    (method_name, method_value)
        2-tuple of the name and value of each non-builtin method bound to this
        object (in ascending lexicographic order of method name).

    See Also
    ----------
    :func:`iter_methods`
        Further details.
    '''

    yield from iter_methods_matching(
        obj=obj, predicate=lambda method_name: not (
            method_name.startswith('__') and method_name.endswith('__')))

# ....................{ ITERATORS ~ vars                   }....................
#FIXME: Reimplement in terms of iter_attrs_matching().
def iter_vars(obj: object) -> GeneratorType:
    '''
    Generator yielding a 2-tuple of the name and value of each variable bound to
    the passed object (in ascending lexicographic order of variable name).

    This includes:

    * All variables statically registered in this object's internal dictionary
      (e.g., ``__dict__`` in unslotted objects), including both:
      * Builtin variables, whose names are both prefixed and suffixed by ``__``.
      * Custom variables, whose names are *not* prefixed and suffixed by ``__``.
    * All variables dynamically returned by **property methods** (i.e., methods
      decorated by the :func:`property` decorator).

    This excludes *only* variables dynamically defined by this object's
    ``__getattr__()`` method or related runtime magic, which cannot (by
    definition) be inspected by this generator.

    Parameters
    ----------
    obj : object
        Object to iterate the variables of.

    Yields
    ----------
    (var_name, var_value)
        2-tuple of the name and value of each variable bound to this object (in
        ascending lexicographic order of variable name).

    Raises
    ----------
    Exception
        If any variable of this object is a property whose
        :func:`property`-decorated method raises an exception. To avoid handling
        such exceptions, consider instead calling the
        :func:`iter_vars_custom_simple` generator excluding all properties.
    '''

    # Return a generator comprehension...
    return (
        # Yielding this variable's name and definition...
        (var_name, var_value)
        # For the name and value of each attribute of this object...
        for var_name, var_value in inspect.getmembers(obj)
        # If this attribute is *NOT* a method, yield this name and value.
        if not callable(var_value)
    )


#FIXME: Reimplement in terms of iter_attrs_matching().
def iter_vars_custom(obj: object) -> GeneratorType:
    '''
    Generator yielding 2-tuples of the name and value of each **non-builtin
    variable** (i.e., variable whose name is *not* both prefixed and suffixed by
    `__`) bound to the passed object (in ascending lexicographic order of
    variable name).

    Parameters
    ----------
    obj : object
        Object to yield all non-builtin variables of.

    Yields
    ----------
    (var_name, var_value)
        2-tuple of the name and value of each non-builtin variable bound to this
        object (in ascending lexicographic order of variable name).

    See Also
    ----------
    :func:`iter_vars`
        Further details.
    '''

    # For the name and value of each variable of this object...
    for var_name, var_value in iter_vars(obj):
        # If this variable is *NOT* builtin, yield this name and value.
        if not (var_name.startswith('__') and var_name.endswith('__')):
            yield var_name, var_value

# ....................{ ITERATORS ~ vars : custom simple   }....................
def iter_vars_custom_simple(obj: object) -> GeneratorType:
    '''
    Generator yielding 2-tuples of the name and value of each **non-builtin
    non-property variable** (i.e., variable whose name is *not* both prefixed
    and suffixed by ``__`` and whose value is *not* dynamically defined by the
    ``@property`` decorator to be the implicit result of a method call) bound to
    the passed object (in ascending lexicographic order of variable name).

    Only variables statically registered in this object's internal dictionary
    (e.g., ``__dict__`` in unslotted objects) are yielded. Variables dynamically
    defined by this object's ``__getattr__()`` method or related runtime magic
    are simply ignored.

    Parameters
    ----------
    obj : object
        Object to yield all non-builtin non-property variables of.

    Yields
    ----------
    (var_name, var_value)
        2-tuple of the name and value of each non-builtin non-property variable
        bound to this object (in ascending lexicographic order of variable
        name).
    '''

    # Defer to this iterator with the null predicate unconditionally accepting
    # all such variables regardless of name or value.
    yield from iter_vars_custom_simple_matching(
        obj=obj, predicate=noop_attr_predicate)


def iter_vars_custom_simple_prefixed(obj: object, prefix: str) -> GeneratorType:
    '''
    Generator yielding 2-tuples of the name and value of each **non-builtin
    non-property prefixed variable** (i.e., variable whose name is prefixed by
    the passed substring and *not* both prefixed and suffixed by ``__`` and
    whose value is *not* dynamically defined by the ``@property`` decorator to
    be the implicit result of a method call) bound to the passed object (in
    ascending lexicographic order of variable name).

    Only variables statically registered in this object's internal dictionary
    (e.g., ``__dict__`` in unslotted objects) are yielded. Variables dynamically
    defined by this object's ``__getattr__()`` method or related runtime magic
    are simply ignored.

    Parameters
    ----------
    obj : object
        Object to yield all non-builtin non-property prefixed variables of.
    prefix : str
        Substring prefixing the names of all such variables.

    Yields
    ----------
    (var_name, var_value)
        2-tuple of the name and value of each non-builtin non-property prefixed
        variable bound to this object (in ascending lexicographic order of
        variable name).
    '''

    # Defer to this iterator.
    yield from iter_vars_custom_simple_matching(
        obj=obj, predicate=lambda var_name, var_value: (
            var_name.startswith(prefix)))


@type_check
def iter_vars_custom_simple_matching(
    obj: object, predicate: CallableTypes) -> GeneratorType:
    '''
    Generator yielding 2-tuples of the name and value of each **non-builtin
    non-property variable** (i.e., variable whose name is *not* both prefixed
    and suffixed by ``__`` and whose value is *not* dynamically defined by the
    ``@property`` decorator to be the implicit result of a method call) bound to
    the passed object whose name and/or value also matches the passed predicate
    (in ascending lexicographic order of variable name).

    Only variables statically registered in this object's internal dictionary
    (e.g., ``__dict__`` in unslotted objects) are yielded. Variables dynamically
    defined by this object's ``__getattr__()`` method or related runtime magic
    are simply ignored.

    Parameters
    ----------
    obj : object
        Object to yield all non-builtin non-property variables of.
    predicate : CallableTypes
        Callable iteratively passed both the name and value of each attribute
        bound to this object, returning ``True`` only if that name and/or value
        matches this predicate.

    Yields
    ----------
    (var_name, var_value)
        2-tuple of the name and value of each matching non-builtin non-property
        variable bound to this object (in ascending lexicographic order of
        variable name).
    '''

    # Ideally, this function would be reimplemented in terms of the
    # iter_attrs_matching() function calling the canonical inspect.getmembers()
    # function. Dynamic inspection is surprisingly non-trivial in the general
    # case, particularly when virtual base classes rear their diamond-studded
    # faces. Moreover, doing so would support edge-case attributes when passed
    # class objects, including:
    #
    # * Metaclass attributes of the passed class.
    # * Attributes decorated by "@DynamicClassAttribute" of the passed class.
    #
    # Sadly, inspect.getmembers() internally accesses attributes via the
    # dangerous getattr() builtin rather than the safe inspect.getattr_static()
    # function. This function explicitly requires the latter and hence *MUST*
    # reimplement rather than defer to inspect.getmembers(). (Sadness reigns.)
    #
    # For the same reason, the unsafe vars() builtin cannot be called either.
    # Since that builtin fails for builtin containers (e.g., "dict", "list"),
    # this is not altogether a bad thing.
    for attr_name in dir(obj):
        # If this attribute is *NOT* a builtin...
        if not (attr_name.startswith('__') and attr_name.endswith('__')):
            # Value of this attribute guaranteed to be statically rather than
            # dynamically retrieved. The getattr() builtin performs the latter,
            # dynamically calling this attribute's getter if this attribute is
            # a property. Since that call could conceivably raise unwanted
            # exceptions *AND* since this function explicitly ignores
            # properties, static attribute retrievable is unavoidable.
            attr_value = inspect.getattr_static(obj, attr_name)

            # If all of the following conditions apply:
            if (
                # This value is neither callable nor a property, this attribute
                # must necessarily be a non-property variable.
                not (
                    callable(attr_value) or
                    isinstance(attr_value, property)
                ) and
                # This non-property variable matches this predicate. For safety,
                # this predicate is passed this attribute only *AFTER* this
                # attribute has been guaranteed to be such a variable; this
                # predicate may thus safely assume this to be the case.
                predicate(attr_name, attr_value)
            # ...yield this matching variable.
            ):
                yield attr_name, attr_value

# ....................{ PREDICATES                         }....................
def noop_attr_predicate(attr_name: object, attr_value: object) -> bool:
    '''
    Null attribute predicate unconditionally returning ``True`` for all passed
    attribute name and value combinations, suitable for passing as the value of
    the ``predicate`` parameter accepted by attribute iterators in this
    submodule (e.g., :func:`iter_attrs_matching`).
    '''

    # ...our work is done here, folks.
    return True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2017 by Alexis Pietak & Cecil Curry
# See "LICENSE" for further details.


import matplotlib.pyplot as plt
from betse.exceptions import BetseExceptionSimulation
from betse.science import filehandling as fh
from betse.science.parameters import Parameters
from betse.science.plot.pipeline import plot_all
from betse.science.sim import Simulator
from betse.util.io.log import logs
from betse.util.path import files, paths
from betse import ignition, pathtree

# Initialize the current application before doing anything else.
ignition.init()
# at the moment only the default config file can be used
config_filename = pathtree.CONFIG_DEFAULT_FILENAME

# Validate and localize such filename.
files.die_unless_file(config_filename)
_config_filename = config_filename
_config_basename = paths.get_basename(_config_filename)

logs.log_info(
    'Plotting simulation with configuration "{}".'.format(_config_basename))

p = Parameters(config_filename=_config_filename)  # create an instance of Parameters
sim = Simulator(p)  # create an instance of Simulator

if files.is_file(sim.savedSim):
    sim, cells, _ = fh.loadSim(sim.savedSim)  # load the simulation from cache
else:
    raise BetseExceptionSimulation(
        "Ooops! No such simulation file found to plot!")

plot_all(cells, sim, p, plot_type='sim')

if p.turn_all_plots_off is False:
    plt.show()


#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# FIXME include other channels in morphogen (dye) dynamics

# ....................{ IMPORTS                            }....................
import copy
import numpy as np
from betse.exceptions import BetseSimTissueException
from betse.science.config.confenum import CellsPickerType, SolverType
from betse.science.math import modulate as mod
from betse.science.math import toolbox as tb
from betse.science.tissue.tisprofile import CutProfile, TissueProfile
from betse.science.tissue.event.tisevecut import SimEventCut
from betse.science.tissue.picker.tispickcls import (
    TissuePickerAll, TissuePickerIndices, TissuePickerPercent)
from betse.science.tissue.picker.tispickimage import TissuePickerImage
from betse.util.io.log import logs
# from betse.util.type import types
from betse.util.type.types import type_check
from collections import OrderedDict
from random import shuffle

# ....................{ CLASSES                            }....................
#FIXME: Since this class handles both tissue profiles *AND* scheduled
#interventions, consider renaming this class to a less ambiguous name -- say, to
#"TissueEventHandler".
#FIXME: Document all instance variables defined within this class.
class TissueHandler(object):
    '''
    High-level tissue handler, managing all tissue-centric functionality
    including both tissue and cut profiles *and* scheduled interventions.

    This handler governs all:

    * Tissue profiles and objects required by these profiles, including all
      geometry-specifying bitmaps.
    * Scheduled interventions, even those *not* pertaining to tissue profiles
      (e.g., global scheduled interventions).

    Attributes (Profile: Tissue)
    ----------
    cell_target_inds : dict
        Dictionary mapping from the name of each tissue profile enabled by the
        current simulation configuration to a one-dimensional NumPy array of the
        indices of all cells in the cluster belonging to this tissue.
    tissue_default : TissueProfile
        **Default tissue profile** (i.e., profile applicable to all cells *not*
        already assigned to a non-default tissue profile). For convenience, this
        profile is accessible via both this instance variable *and* as a
        standard tissue profile (i.e., as a value of a key-value pair of the
        ordered :attr:`tissue_name_to_profile` dictionary whose key is the name
        of this profile).
    tissue_name_to_profile : OrderedDict
        Ordered dictionary mapping from the name of each tissue profile enabled
        by the current simulation configuration (in YAML list order) to the
        :class:`TissueProfile` instance encapsulating this profile.
    tissue_target_inds : dict
        Dictionary mapping from the name of each tissue profile enabled by the
        current simulation configuration to a one-dimensional Numpy array of the
        indices of all cell membranes in the cluster belonging to this tissue.

    Attributes (Profile: Cut)
    ----------
    cut_name_to_profile : OrderedDict
        Ordered dictionary mapping from the name of each cut profile enabled by
        the current simulation configuration (in YAML list order) to the
        :class:`CutProfile` instance encapsulating this profile.
    event_cut : SimEventCut
        **Cutting event** (i.e., event removing a region of the current cluster
        at some time step during the simulation phase) if enabled by the
        current simulation configuration *or* ``None`` otherwise.
    _wound_channel : {TRP, NoneType}
        Wound-induced TRP-based channel if the following conditions are all
        satisfied *or* ``None`` otherwise:
        * The current simulation configuration has enabled simulation of the
          mechanosensitive Na channel associated with wounding.
        * The cutting event has already been performed, thus generating a wound
          and activating this channel.
    '''

    # ..................{ INITIALIZERS                       }..................
    # Avoid circular import dependencies.
    @type_check
    def __init__(self, p: 'betse.science.parameters.Parameters') -> None:
        '''
        Initialize this tissue handler.

        Parameters
        ----------
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        '''

        # Initialize tissue and cut profiles.
        self._init_tissues(p)
        self._init_cuts(p)

        # Initialize scheduled interventions *AFTER* tissue and cut profiles, as
        # the former requires the latter.
        self._init_events(p)


    @type_check
    def _init_tissues(self, p: 'betse.science.parameters.Parameters') -> None:
        '''
        Internally initialize all tissue profiles defined by the passed
        simulation configuration.

        Specifically, this method encapsulates each low-level YAML-formatted
        list item defining a tissue profile in this configuration with a
        high-level instance of the :class:`CellsProfileABC` class.
        '''

        # Ordered dictionaries mapping from the name of each each tissue profile
        # to the corresponding profile.
        self.tissue_name_to_profile = OrderedDict()

        # Nullify all remaining tissue-centric attributes for safety.
        self.cell_target_inds = None
        self.env_target_inds = None
        self.tissue_target_inds = None

        # If tissue profiles are disabled, silently noop.
        if not p.is_tissue_profiles:
            return

        # 1-based index of the default tissue profile.
        tissue_z_order = 1

        # Object assigning a cell cluster region to the default tissue profile.
        tissue_picker = TissuePickerAll()

        #FIXME: Set this picker for the "tissue_default" object, while still
        #setting the above tissue picker for the
        #"tissue_name_to_profile["p.tissue_default.name]" object -- implying
        #that these two objects should be different objects. Viola!
        # tissue_picker = TissuePickerImage(
        #     filename=p.tissue_default.picker_image_filename,
        #     dirname=p.conf_dirname)

        # For convenience, encapsulate the default tissue profile as a
        # high-level object accessible both as a separate instance variable
        # *AND* as a standard tissue profile (i.e., as a value of a key-value
        # pair whose key is the name of the default tissue profile).
        self.tissue_default = self.tissue_name_to_profile[
            p.tissue_default.name] = TissueProfile(
                name=p.tissue_default.name,
                z_order=tissue_z_order,
                picker=tissue_picker,

                # By definition, the cell cluster shares no gap junctions with cells
                # in the environment -- since, of course, there *ARE* no cells
                # in the environment.
                is_gj_insular=True,

                Dm_Na=p.tissue_default.Dm_Na,
                Dm_K=p.tissue_default.Dm_K,
                Dm_Cl=p.tissue_default.Dm_Cl,
                Dm_Ca=p.tissue_default.Dm_Ca,
                Dm_M=p.tissue_default.Dm_M,
                Dm_P=p.tissue_default.Dm_P,
            )

        # For each non-default tissue profile...
        for tissue_profile in p.tissue_profiles:
            # If a prior profile collides with this profile's name, this profile
            # is non-unique. In this case, raise an exception.
            if tissue_profile.name in self.tissue_name_to_profile:
                raise BetseSimTissueException(
                    'Tissue profile "{0}" non-unique '
                    '(i.e., two or more tissue profiles named "{0}").'.format(
                        tissue_profile.name))

            # 1-based index of this tissue profile.
            tissue_z_order += 1

            # Object assigning a cell cluster region to this tissue profile.
            tissue_picker = None

            # Conditionally define this object.
            if tissue_profile.picker_type is CellsPickerType.ALL:
                tissue_picker = TissuePickerAll()
            elif tissue_profile.picker_type is CellsPickerType.IMAGE:
                tissue_picker = TissuePickerImage(
                    filename=tissue_profile.picker_image_filename,
                    dirname=p.conf_dirname)
            elif tissue_profile.picker_type is CellsPickerType.INDICES:
                tissue_picker = TissuePickerIndices(
                    cells_index=tissue_profile.picker_cells_index)
            elif tissue_profile.picker_type is CellsPickerType.PERCENT:
                tissue_picker = TissuePickerPercent(
                    cells_percent=tissue_profile.picker_cells_percent)
            else:
                raise BetseSimTissueException(
                    'Tissue profile picker type "{}" unrecognized.'.format(
                        tissue_profile.picker_type))

            # Map this profile's name to a high-level tissue profile object.
            self.tissue_name_to_profile[tissue_profile.name] = TissueProfile(
                name=tissue_profile.name,
                z_order=tissue_z_order,
                picker=tissue_picker,
                is_gj_insular=tissue_profile.is_gj_insular,
                Dm_Na=tissue_profile.Dm_Na,
                Dm_K=tissue_profile.Dm_K,
                Dm_Cl=tissue_profile.Dm_Cl,
                Dm_Ca=tissue_profile.Dm_Ca,
                Dm_M=tissue_profile.Dm_M,
                Dm_P=tissue_profile.Dm_P,
            )


    @type_check
    def _init_cuts(self, p: 'betse.science.parameters.Parameters') -> None:
        '''
        Internally initialize all cut profiles defined by the passed simulation
        configuration.

        Specifically, this method encapsulates each low-level YAML-formatted
        list item defining a cut profile in this configuration with a high-level
        instance of the :class:`CellsProfileABC` class.
        '''

        # Ordered dictionaries mapping from the name of each each cut profile
        # to the corresponding profile.
        self.cut_name_to_profile = OrderedDict()

        # If cut profiles are disabled, silently noop.
        if not p.is_tissue_profiles:
            return

        # For each low-level cut profile...
        for cut_index, cut_profile in enumerate(p.cut_profiles):
            # If this profile is non-unique, raise an exception.
            if cut_profile.name in self.cut_name_to_profile:
                raise BetseSimTissueException(
                    'Cut profile "{0}" non-unique '
                    '(i.e., two or more cut profiles named "{0}").'.format(
                        cut_profile.name))

            # Map this profile's name to a high-level profile object.
            self.cut_name_to_profile[cut_profile.name] = CutProfile(
                name=cut_profile.name,
                z_order=cut_index + 1,
                picker=TissuePickerImage(
                    filename=cut_profile.picker_image_filename,
                    dirname=p.conf_dirname)
            )


    @type_check
    def _init_events(self, p: 'betse.science.parameters.Parameters') -> None:
        '''
        Initialize all scheduled interventions defined by the passed simulation
        configuration.
        '''

        # Cutting event if enabled by this configuration or "None" otherwise.
        self.event_cut = None

        # Wound-induced TRP-based channel if enabled by this configuration and a
        # cutting event generating such a wound has already been performed or
        # "None" otherwise.
        self._wound_channel = None

        #FIXME: Access only public "Parameters" attributes.
        # If this event is enabled...
        ce = p._conf['cutting event']
        if bool(ce['event happens']):
            # If cut profiles are enabled...
            if self.cut_name_to_profile:
                # List of the names of all cut profiles performed by this event.
                cut_profile_names = ce['apply to']

                # For each such name...
                for cut_profile_name in cut_profile_names:
                    # If this cut profile does *NOT* exist, raise an exception.
                    if cut_profile_name not in self.cut_name_to_profile:
                        raise BetseSimTissueException(
                            'Cut profile "{}" referenced by '
                            'cutting event not found.'.format(
                                cut_profile_name))

                # Define this event.
                self.event_cut = SimEventCut(
                    # Time step at which to cut. For simplicity, this is coerced
                    # to be the start of the simulation.
                    time_step=0.0,
                    profile_names=cut_profile_names,
                )
            # Else, log a non-fatal warning.
            else:
                logs.log_warning(
                    'Ignoring cutting event, as cut profiles are disabled.')


    #FIXME: Rename to _map_tissue_profiles_to_cells(). See the
    #SimPhase.__init__() method for further discussion.
    @type_check
    def tissueProfiles(
        self,
        sim:   'betse.science.sim.Simulator',
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> None:
        '''
        Define all dictionaries mapping from tissue profile names to
        one-dimensional NumPy arrays of the indices of various cellular objects
        (e.g., cells, cell membranes) in the cluster belonging to these tissues.

        Specifically, this method defines the :attr:`cell_target_inds`,
        :attr:`env_target_inds`, and :attr:`tissue_target_inds` dictionaries.
        '''

        # Log this initialization.
        logs.log_debug('Mapping tissue, cut, and boundary profiles...')

        #FIXME: These three dictionaries are all indexed by tissue profile names
        #and hence should simply be folded into the "TissueProfile" class as
        #instance variables whose values are the corresponding one-dimensional
        #Numpy lists for that specific tissue. This eliminates awkward lookups.
        self.cell_target_inds = {}
        self.env_target_inds = {}

        #FIXME: Let's rename this to something less ambiguous. Since it indexes
        #cell membranes, "mem_target_inds" might be a reasonable name? Lo-ho-ho!
        self.tissue_target_inds = {}

        # For each tissue profiles defined by this simulation configuration...
        for tissue_name, tissue_profile in self.tissue_name_to_profile.items():
            # One-dimensional Numpy arrays of the indices of all cells and cell
            # membranes comprising this tissue.
            tissue_cells_index, tissue_mems_index = (
                tissue_profile.picker.pick_cells_and_mems(cells=cells, p=p))

            # Persist these arrays for this tissue.
            self.cell_target_inds[tissue_name] = tissue_cells_index
            self.tissue_target_inds[tissue_name] = tissue_mems_index

            # If this tissue is non-empty (i.e., contains at least one cell)...
            if len(self.cell_target_inds[tissue_name]):
                # Get ECM targets.
                if p.is_ecm:
                    ecm_targs_mem = list(cells.map_mem2ecm[tissue_mems_index])
                    self.env_target_inds[tissue_name] = ecm_targs_mem

                # Set the values of Dmems and ECM diffusion based on the
                # identified target indices.
                if p.ions_dict['Na'] == 1:
                    sim.Dm_cells[sim.iNa][tissue_mems_index] = (
                        tissue_profile.Dm_Na)

                if p.ions_dict['K'] == 1:
                    sim.Dm_cells[sim.iK][tissue_mems_index] = (
                        tissue_profile.Dm_K)

                if p.ions_dict['Cl'] == 1:
                    sim.Dm_cells[sim.iCl][tissue_mems_index] = (
                        tissue_profile.Dm_Cl)

                if p.ions_dict['Ca'] == 1:
                    sim.Dm_cells[sim.iCa][tissue_mems_index] = (
                        tissue_profile.Dm_Ca)

                if p.ions_dict['M'] == 1:
                    sim.Dm_cells[sim.iM][tissue_mems_index] = (
                        tissue_profile.Dm_M)

                if p.ions_dict['P'] == 1:
                    sim.Dm_cells[sim.iP][tissue_mems_index] = (
                        tissue_profile.Dm_P)

    # ..................{ RUNNERS ~ init                     }..................
    def runAllInit(
        self, sim: 'Simulator', cells: 'Cells', p: 'Parameters') -> None:
        '''
        Initialize all tissue manipulations specified by the passed
        user-specified parameters with the passed tissue simulation and
        cellular world.
        '''

        self._init_events_global(  sim, cells, p)
        self._init_events_tissue(  sim, cells, p)

    # ..................{ RUNNERS ~ apply                    }..................
    def runAllDynamics(self, sim, cells, p, t: float):
        '''
        Apply all tissue manipulations specified by the passed user-specified
        parameters to the passed tissue simulation and cellular world for the
        passed time step.
        '''

        self._sim_events_global(  sim, cells, p, t)
        self._sim_events_tissue(  sim, cells, p, t)
        self.makeAllChanges(sim)

    def _init_events_global(self,sim,cells,p):
        '''
        Initialize all **global scheduled interventions** (i.e., events globally
        applicable to all cells) specified by the passed user-specified
        parameters with the passed tissue simulation and cellular world.
        '''

        if p.global_options['K_env'] != 0:

            self.t_on_Kenv = p.global_options['K_env'][0]
            self.t_off_Kenv = p.global_options['K_env'][1]
            self.t_change_Kenv = p.global_options['K_env'][2]
            self.mem_mult_Kenv = p.global_options['K_env'][3]

        if p.global_options['Cl_env'] != 0:

            self.t_on_Clenv = p.global_options['Cl_env'][0]
            self.t_off_Clenv = p.global_options['Cl_env'][1]
            self.t_change_Clenv = p.global_options['Cl_env'][2]
            self.mem_mult_Clenv = p.global_options['Cl_env'][3]

        if p.global_options['Na_env'] != 0:

            self.t_on_Naenv = p.global_options['Na_env'][0]
            self.t_off_Naenv = p.global_options['Na_env'][1]
            self.t_change_Naenv = p.global_options['Na_env'][2]
            self.mem_mult_Naenv = p.global_options['Na_env'][3]

        if p.global_options['T_change'] != 0:

            self.tonT = p.global_options['T_change'][0]
            self.toffT = p.global_options['T_change'][1]
            self.trampT = p.global_options['T_change'][2]
            self.multT = p.global_options['T_change'][3]

        if p.global_options['gj_block'] != 0:

            self.tonGJ = p.global_options['gj_block'][0]
            self.toffGJ = p.global_options['gj_block'][1]
            self.trampGJ = p.global_options['gj_block'][2]

            numo = p.global_options['gj_block'][3]

            numo = int(numo)

            if numo > 100:
                numo = 100
            elif numo < 1:
                numo = 1

            #FIXME: Isn't the exact same "data_length" already available as the
            #"self.data_length" variable initialized by the __init__() method?
            #Leaping unicorns deny the inevitability of love!
            data_length = len(cells.mem_i)
            data_fraction = int((numo/100)*data_length)

            mem_i_copy = np.copy(cells.mem_i[:])
            shuffle(mem_i_copy)
            self.targets_gj_block = [cells.mem_i[x] for x in range(0,data_fraction)]

        if p.global_options['NaKATP_block'] != 0:
            self.tonNK = p.global_options['NaKATP_block'][0]
            self.toffNK = p.global_options['NaKATP_block'][1]
            self.trampNK = p.global_options['NaKATP_block'][2]

    def _init_events_tissue(self, sim, cells, p):
        '''
        Initialize all **targeted scheduled interventions** (i.e., events only
        applicable to specific tissue profiles) specified by the passed
        user-specified parameters with the passed tissue simulation and cellular
        world.
        '''

        if p.scheduled_options['Na_mem'] != 0:
            self.t_on_Namem = p.scheduled_options['Na_mem'][0]
            self.t_off_Namem = p.scheduled_options['Na_mem'][1]
            self.t_change_Namem = p.scheduled_options['Na_mem'][2]
            self.mem_mult_Namem = p.scheduled_options['Na_mem'][3]
            self.apply_Namem = p.scheduled_options['Na_mem'][4]
            self.function_Namem = p.scheduled_options['Na_mem'][5]

            self.targets_Namem = []
            for profile in self.apply_Namem:
                targets = self.tissue_target_inds[profile]
                self.targets_Namem.append(targets)

            self.targets_Namem = [
                item for sublist in self.targets_Namem for item in sublist]

            self.scalar_Namem = 1
            self.dyna_Namem = lambda t: 1

            # call a special toolbox function to change membrane permeability: spatial grads
            # 'gradient_x', 'gradient_y', 'gradient_r'

            if self.function_Namem != 'None':
                self.scalar_Namem, self.dyna_Namem = getattr(
                    mod, self.function_Namem)(self.targets_Namem,cells,p)

        #----------------------------------------------

        if p.scheduled_options['K_mem'] != 0:
            self.t_on_Kmem = p.scheduled_options['K_mem'][0]
            self.t_off_Kmem = p.scheduled_options['K_mem'][1]
            self.t_change_Kmem = p.scheduled_options['K_mem'][2]
            self.mem_mult_Kmem = p.scheduled_options['K_mem'][3]
            self.apply_Kmem = p.scheduled_options['K_mem'][4]
            self.function_Kmem = p.scheduled_options['K_mem'][5]

            self.targets_Kmem = []
            for profile in self.apply_Kmem:
                targets = self.tissue_target_inds[profile]
                self.targets_Kmem.append(targets)

            self.targets_Kmem = [
                item for sublist in self.targets_Kmem for item in sublist]

            self.scalar_Kmem = 1
            self.dyna_Kmem = lambda t: 1

            if self.function_Kmem != 'None':
                # call a special toolbox function to change membrane permeability: spatial grads
                # 'gradient_x', 'gradient_y', 'gradient_r'

                self.scalar_Kmem, self.dyna_Kmem = getattr(
                    mod, self.function_Kmem)(self.targets_Kmem, cells, p)

        #----------------------------------------------

        if p.scheduled_options['Cl_mem'] != 0:
            self.t_on_Clmem = p.scheduled_options['Cl_mem'][0]
            self.t_off_Clmem = p.scheduled_options['Cl_mem'][1]
            self.t_change_Clmem = p.scheduled_options['Cl_mem'][2]
            self.mem_mult_Clmem = p.scheduled_options['Cl_mem'][3]
            self.apply_Clmem = p.scheduled_options['Cl_mem'][4]
            self.function_Clmem = p.scheduled_options['Cl_mem'][5]

            self.targets_Clmem = []
            for profile in self.apply_Clmem:
                targets = self.tissue_target_inds[profile]
                self.targets_Clmem.append(targets)

            self.targets_Clmem = [
                item for sublist in self.targets_Clmem for item in sublist]

            self.scalar_Clmem = 1
            self.dyna_Clmem = lambda t: 1

            if self.function_Clmem != 'None':

                # call a special toolbox function to change membrane permeability: spatial grads
                # 'gradient_x', 'gradient_y', 'gradient_r'
                self.scalar_Clmem, self.dyna_Clmem = getattr(mod,self.function_Clmem)(self.targets_Clmem,cells,p)

        #----------------------------------------------

        if p.scheduled_options['Ca_mem'] != 0:
            self.t_on_Camem = p.scheduled_options['Ca_mem'][0]
            self.t_off_Camem = p.scheduled_options['Ca_mem'][1]
            self.t_change_Camem = p.scheduled_options['Ca_mem'][2]
            self.mem_mult_Camem = p.scheduled_options['Ca_mem'][3]
            self.apply_Camem = p.scheduled_options['Ca_mem'][4]
            self.function_Camem = p.scheduled_options['Ca_mem'][5]

            self.targets_Camem = []
            for profile in self.apply_Camem:
                targets = self.tissue_target_inds[profile]
                self.targets_Camem.append(targets)

            self.targets_Camem = [
                item for sublist in self.targets_Camem for item in sublist]

            self.scalar_Camem = 1
            self.dyna_Camem = lambda t: 1

            if self.function_Camem != 'None':
                # call a special toolbox function to change membrane permeability: spatial grads
                # 'gradient_x', 'gradient_y', 'gradient_r'

                self.scalar_Camem, self.dyna_Camem = getattr(mod, self.function_Camem)(self.targets_Camem,cells,p)

        #----------------------------------------------

        if p.scheduled_options['pressure'] != 0:

            self.t_onP = p.scheduled_options['pressure'][0]
            self.t_offP = p.scheduled_options['pressure'][1]
            self.t_changeP = p.scheduled_options['pressure'][2]
            self.rate_P = p.scheduled_options['pressure'][3]
            self.apply_P = p.scheduled_options['pressure'][4]
            self.function_P = p.scheduled_options['pressure'][5]

            self.targets_P = []
            for profile in self.apply_P:
                targets = self.cell_target_inds[profile]
                self.targets_P.append(targets)

            self.targets_P = [
                item for sublist in self.targets_P for item in sublist]

            self.scalar_P = 1
            self.dyna_P = lambda t: 1

            # call a special toolbox function to change membrane permeability: spatial grads
            # 'gradient_x', 'gradient_y', 'gradient_r'
            if self.function_P != 'None':
                self.scalar_P, self.dyna_P = getattr(mod, self.function_P)(
                    self.targets_P,cells, p)

        #--------------------------------------------------------

        if p.scheduled_options['ecmJ'] != 0 and p.is_ecm:
            self.t_on_ecmJ  = p.scheduled_options['ecmJ'][0]
            self.t_off_ecmJ = p.scheduled_options['ecmJ'][1]
            self.t_change_ecmJ = p.scheduled_options['ecmJ'][2]
            self.apply_ecmJ = p.scheduled_options['ecmJ'][3]
            self.mult_ecmJ = p.scheduled_options['ecmJ'][4]

            self.targets_ecmJ = []
            for profile in self.apply_ecmJ:
                targets = self.env_target_inds[profile]
                self.targets_ecmJ.append(targets)

            self.targets_ecmJ = [
                item for sublist in self.targets_ecmJ for item in sublist]

    def _sim_events_global(self, sim, cells, p, t) -> None:
        '''
        Apply all **global scheduled interventions** (i.e., events globally
        applicable to all cells) specified by the passed user-specified
        parameters to the passed tissue simulation and cellular world for the
        passed time step.
        '''

        if p.global_options['K_env'] != 0:
            effector_Kenv = tb.pulse(
                t, self.t_on_Kenv, self.t_off_Kenv, self.t_change_Kenv)

            if p.is_ecm: # simulate addition of potassium salt to remain charge neutral
                sim.c_env_bound[sim.iK] = (
                    self.mem_mult_Kenv*effector_Kenv*p.env_concs['K'] +
                    p.env_concs['K']
                )
                sim.c_env_bound[sim.iM] = (
                    self.mem_mult_Kenv*effector_Kenv*p.env_concs['K'] +
                    p.env_concs['M']
                )
            else:
                sim.cc_env[sim.iK][:] = (
                    self.mem_mult_Kenv*effector_Kenv*p.conc_env_k + p.conc_env_k)

        if p.global_options['Cl_env'] != 0 and p.ions_dict['Cl'] == 1:
            effector_Clenv = tb.pulse(t,self.t_on_Clenv,self.t_off_Clenv,self.t_change_Clenv)

            if not p.is_ecm:
                sim.cc_env[sim.iCl][:] = self.mem_mult_Clenv*effector_Clenv*p.conc_env_cl + p.conc_env_cl

            else:  # simulate addition of sodium chloride to remain charge neutral
                sim.c_env_bound[sim.iCl] = self.mem_mult_Clenv*effector_Clenv*p.env_concs['Cl'] + p.env_concs['Cl']
                sim.c_env_bound[sim.iNa] = self.mem_mult_Clenv*effector_Clenv*p.env_concs['Cl'] + p.env_concs['Na']

        if p.global_options['Na_env'] != 0:
            effector_Naenv = tb.pulse(t,self.t_on_Naenv,self.t_off_Naenv,self.t_change_Naenv)

            if not p.is_ecm:
                sim.cc_env[sim.iNa][:] = self.mem_mult_Naenv*effector_Naenv*p.conc_env_na + p.conc_env_na

            else: # simulate addition of sodium salt to remain charge neutral
                sim.c_env_bound[sim.iNa] = self.mem_mult_Naenv*effector_Naenv*p.env_concs['Na'] + p.env_concs['Na']
                sim.c_env_bound[sim.iM] = self.mem_mult_Naenv*effector_Naenv*p.env_concs['Na'] + p.env_concs['M']

        if p.global_options['T_change'] != 0:
            sim.T = self.multT*tb.pulse(t,self.tonT,self.toffT,self.trampT)*p.T + p.T

        if p.global_options['gj_block'] != 0:
            sim.gj_block[self.targets_gj_block] = (1.0 - tb.pulse(t,self.tonGJ,self.toffGJ,self.trampGJ))

        if p.global_options['NaKATP_block'] != 0:
            sim.NaKATP_block = (1.0 - tb.pulse(t,self.tonNK,self.toffNK,self.trampNK))

    def _sim_events_tissue(self, sim, cells, p, t):
        '''
        Apply all **targeted scheduled interventions** (i.e., events only
        applicable to specific tissue profiles) specified by the passed
        user-specified parameters to the passed tissue simulation and cellular
        world for the passed time step.
        '''

        if p.scheduled_options['Na_mem'] != 0:
            effector_Na = (
                self.scalar_Namem*self.dyna_Namem(t)*
                tb.pulse(t,self.t_on_Namem,self.t_off_Namem,self.t_change_Namem))
            sim.Dm_scheduled[sim.iNa][self.targets_Namem] = (
                self.mem_mult_Namem*effector_Na*self.tissue_default.Dm_Na)

        if p.scheduled_options['K_mem'] != 0:
            effector_K = (
                self.scalar_Kmem*self.dyna_Kmem(t)*
                tb.pulse(t,self.t_on_Kmem,self.t_off_Kmem,self.t_change_Kmem))
            sim.Dm_scheduled[sim.iK][self.targets_Kmem] = (
                self.mem_mult_Kmem*effector_K*self.tissue_default.Dm_K)

        if p.scheduled_options['Cl_mem'] != 0 and p.ions_dict['Cl'] != 0:
            effector_Cl = (
                self.scalar_Clmem*self.dyna_Clmem(t)*
                tb.pulse(t,self.t_on_Clmem,self.t_off_Clmem,self.t_change_Clmem))
            sim.Dm_scheduled[sim.iCl][self.targets_Clmem] = (
                self.mem_mult_Clmem*effector_Cl*self.tissue_default.Dm_Cl)

        if p.scheduled_options['Ca_mem'] != 0 and p.ions_dict['Ca'] != 0:
            effector_Ca = (
                self.scalar_Camem*self.dyna_Camem(t)*
                tb.pulse(t,self.t_on_Camem,self.t_off_Camem,self.t_change_Camem))
            sim.Dm_scheduled[sim.iCa][self.targets_Camem] = (
                self.mem_mult_Camem*effector_Ca*self.tissue_default.Dm_Ca)

        if p.scheduled_options['pressure'] != 0:
            # logs.log_debug('Applying pressure event...')
            sim.P_mod[self.targets_P] = (
                self.scalar_P*self.dyna_P(t)*self.rate_P*tb.pulse(
                    t,self.t_onP, self.t_offP, self.t_changeP))

        if p.scheduled_options['ecmJ'] != 0:
            if p.is_ecm:
                for i, dmat in enumerate(sim.D_env):
                    effector_ecmJ = self.mult_ecmJ*tb.pulse(
                        t,self.t_on_ecmJ,self.t_off_ecmJ,self.t_change_ecmJ)

                    sim.D_env[i][self.targets_ecmJ] = (
                        sim.D_env_base[i][self.targets_ecmJ]*(
                            1 - effector_ecmJ) + effector_ecmJ*sim.D_free[i])

                    sim.D_env_weight = sim.D_env_weight.ravel()
                    sim.D_env_weight_base = sim.D_env_weight_base.ravel()

                    sim.D_env_weight[self.targets_ecmJ] = \
                        sim.D_env_weight_base[self.targets_ecmJ]*(1-effector_ecmJ) + \
                                                          effector_ecmJ

                    sim.D_env_weight = sim.D_env_weight.reshape(cells.X.shape)
                    sim.D_env_weight_base = sim.D_env_weight_base.reshape(cells.X.shape)

            else:

                effector_ecmJ = self.mult_ecmJ * tb.pulse(
                    t, self.t_on_ecmJ, self.t_off_ecmJ, self.t_change_ecmJ)

                sim.D_env_weight = sim.D_env_weight.ravel()
                sim.D_env_weight_base = sim.D_env_weight_base.ravel()

                sim.D_env_weight[self.targets_ecmJ] = \
                    sim.D_env_weight_base[self.targets_ecmJ] * (1 - effector_ecmJ) + \
                    effector_ecmJ

                sim.D_env_weight = sim.D_env_weight.reshape(cells.X.shape)
                sim.D_env_weight_base = sim.D_env_weight_base.reshape(cells.X.shape)

        # If the cutting event is enabled but has yet to be performed, do so.
        if self.event_cut is not None and not self.event_cut.is_fired and t > p.cut_time:
            for cut_profile_name in self.event_cut.profile_names:
                logs.log_info(
                    'Cutting cell cluster via cut profile "%s"...',
                    cut_profile_name)

                cut_profile_picker = (
                    self.cut_name_to_profile[cut_profile_name].picker)
                self.removeCells(cut_profile_picker, sim, cells, p)

            logs.log_info("Cutting event successful! Resuming simulation...")

            # Redo main data length variable for this dynamics module with
            # updated world.
            self.data_length = len(cells.mem_i)

            self.tissueProfiles(sim, cells, p)
            cells.redo_gj(self, p)
            self.runAllInit(sim, cells, p)

            # Avoid repeating this cutting event at subsequent time steps.
            self.event_cut.fire()

        # If the voltage event is enabled, adjust the voltage accordingly.
        if p.scheduled_options['extV'] is not None:
            p.scheduled_options['extV'].fire(sim, t)


    def stretchChannel(self,sim,cells,p,t):

        dd = np.sqrt(sim.d_cells_x**2 + sim.d_cells_y**2)

        # calculate strain from displacement
        eta = (dd/cells.R)

        # # create a smooth bivariate spline to interpolate deformation data from cells:
        # cellinterp_x = SmoothBivariateSpline(cells.cell_centres[:, 0], cells.cell_centres[:, 1], sim.d_cells_x, kx=1,
        #                                      ky=1)
        # cellinterp_y = SmoothBivariateSpline(cells.cell_centres[:, 0], cells.cell_centres[:, 1], sim.d_cells_y, kx=1,
        #                                      ky=1)
        #
        # # calculate deformations wrt the ecm using the smooth bivariate spline:
        # dmem_x = cellinterp_x.ev(cells.mem_mids_flat[:, 0], cells.mem_mids_flat[:, 1])
        # dmem_y = cellinterp_y.ev(cells.mem_mids_flat[:, 0], cells.mem_mids_flat[:, 1])
        #
        # # obtain normal component to membrane
        # dd = dmem_x*cells.mem_vects_flat[:,2] + dmem_y*cells.mem_vects_flat[:,3]
        #
        # # strain is the divergence of the displacement:
        # eta = np.dot(cells.M_sum_mems, dd*cells.mem_sa)/cells.cell_vol

        # self.active_NaStretch[self.targets_NaStretch] = tb.hill(sim.P_cells[cells.mem_to_cells][self.targets_NaStretch],
        #         self.NaStretch_halfmax,self.NaStretch_n)

        self.active_NaStretch[self.targets_NaStretch] = tb.hill(eta[cells.mem_to_cells][self.targets_NaStretch],
                self.NaStretch_halfmax,self.NaStretch_n)

        sim.Dm_stretch[sim.iNa] = self.maxDmNaStretch*self.active_NaStretch
        sim.Dm_stretch[sim.iK] = self.maxDmNaStretch*self.active_NaStretch

    def makeAllChanges(self, sim) -> None:
        '''
        Add together all effects to finalize changes to cell membrane
        permeabilities for the current time step.
        '''

        sim.Dm_cells = (
            sim.Dm_scheduled +
            sim.Dm_base
        )

        sim.P_cells = sim.P_mod + sim.P_base


    @type_check
    def removeCells(
        self,
        tissue_picker,
        sim:   'betse.science.sim.Simulator',
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> None:
        '''
        Permanently remove all cells selected by the passed tissue picker.

        Parameters
        ---------------------------------
        tissue_picker : TissuePickerABC
            Object matching all cells to be removed.
        sim : betse.science.sim.Simulation
            Current simulation.
        cells : betse.science.cells.Cells
            Current cell cluster.
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        '''

        # Redo environmental diffusion matrices by setting the environmental spaces
        # around cut world to the free value (True) or not (False)?
        # open_TJ = True

        # Subtract this bitmap's clipping mask from the global cluster mask.
        # bitmap_mask = tissue_picker.get_image_mask(cells).clipping_matrix
        # cells.cluster_mask = cells.cluster_mask - bitmap_mask

        # FIXME, if deformation is too much, the following line will crash as
        # the "target_inds_cell" is null. Rayse an uman weedable eggseption.

        # Indices of all cells and cell membranes to be removed.
        target_inds_cell, target_inds_mem = tissue_picker.pick_cells_and_mems(
            cells=cells, p=p)
            # dmem_list = tissue_profile['diffusion constants']

        # get the corresponding flags to membrane entities
        target_inds_gj,_,_ = tb.flatten(cells.cell_to_nn_full[target_inds_cell])

        if p.is_ecm:
            # get environmental targets around each removed cell:
            ecm_targs_cell = list(cells.map_cell2ecm[target_inds_cell])
            ecm_targs_mem = list(cells.map_mem2ecm[target_inds_mem])

            ecm_targs = []
            for v in ecm_targs_cell:
                ecm_targs.append(v)
            for v in ecm_targs_mem:
                ecm_targs.append(v)

            # if sim.molecules is not None and 'ATP' in sim.molecules.core.molecules:
            #
            #     # get concentration of ATP in cells to be removed:
            #     cell_ATP = sim.molecules.core.cell_concs['ATP'][target_inds_cell] * cells.cell_vol[target_inds_cell]
            #     # move this entire concentration to the extracellular spaces (assumed upon cell bursting)
            #     sim.molecules.core.env_concs['ATP'][ecm_targs_cell] = cell_ATP / (p.cell_height * cells.delta ** 2)
            #
            # elif sim.grn is not None and 'ATP' in sim.grn.core.molecules:
            #
            #     # get concentration of ATP in cells to be removed:
            #     cell_ATP = sim.grn.core.cell_concs['ATP'][target_inds_cell] * cells.cell_vol[target_inds_cell]
            #     # move this entire concentration to the extracellular spaces (assumed upon cell bursting)
            #     sim.grn.core.env_concs['ATP'][ecm_targs_cell] = cell_ATP / (p.cell_height * cells.delta ** 2)

            # redo environmental diffusion matrices by
            # setting the environmental spaces around cut world to the free value -- if desired!:
            # if open_TJ is True:
        # save the x,y coordinates of the original boundary cell and membrane points:
        # old_bflag_cellxy = np.copy(cells.cell_centres[cells.bflags_cells])
        # old_bflag_memxy = np.copy(cells.mem_mids_flat[cells.bflags_mems])

        # set up the situation to make world joined to cut world have more permeable membranes:
        hurt_cells = np.zeros(len(cells.cell_i))

        # If we're creating a wound-induced channel ------------------------------------------------
        target_inds_gj_unique = np.unique(target_inds_gj)

        for i, inds in enumerate(cells.cell_to_nn_full): # for all the nn inds to a cell...
            inds_array = np.asarray(inds)
            inds_in_target = np.intersect1d(inds_array,target_inds_gj_unique)

            if len(inds_in_target):
                hurt_cells[i] = 1  # flag the cell as a "hurt" cell

        hurt_inds = (hurt_cells == 1).nonzero()
        sim.hurt_mask = np.zeros(sim.cdl)
        sim.hurt_mask[hurt_inds] = 1.0

        #----------------------------------------------------------------------------------

        # Names of all attributes in the current "Simulation" object.
        sim_names = list(sim.__dict__.keys())

        #FIXME: Redeclare as a "set" object, rename to "specials_names", and remove
        #the duplicate declaration of that variable below. Jumpin' jallopies!

        # Names of all such attributes to be repaired due to this cutting event.
        specials_list = [
            'cc_cells',
            'cc_mems',
            'cc_env',
            'cc_er',
            'z_array',
            'z_array_er',
            'fluxes_gj',
            'fluxes_mem',
            'Dm_base',
            'Dm_cells',
            'Dm_scheduled',
            'Dm_vg',
            'Dm_cag',
            'Dm_morpho',
            'Dm_er_base',
            'Dm_er_CICR',
            'Dm_stretch',
            'Dm_custom',
            'D_gj',
            'fluxes_intra',
            'cc_at_mem',
        ]

        if p.is_ecm:
            specials_list.remove('cc_env')
            extra = ['z_array_cells']
            for ent in extra:
                specials_list.append(ent)

        special_names = set(specials_list)

        #FIXME: This is awesome-sauce in a BETSE jar, but I don't quite grok what's
        #going on. It'd be swell if this could recieve some documentation
        #massaging. The tidal waves of time recede, inch by minute inch!
        for name in sim_names:
            if name in special_names: # if this is a nested data structure...
                super_data = getattr(sim, name)
                super_data2 = []

                for i, data in enumerate(super_data):
                    if isinstance(data, np.ndarray):
                        if len(data) == len(cells.cell_i):
                            data2 = np.delete(data,target_inds_cell)

                        elif len(data) == len(cells.mem_i):
                            data2 = np.delete(data,target_inds_mem)

                        # elif len(data) == len(cells.nn_i):
                        #     data2 = np.delete(data,target_inds_gj)

                    elif isinstance(data, list):
                        data2 = []
                        if len(data) == len(cells.cell_i):
                            for index in sorted(target_inds_cell, reverse=True):
                                del data[index]
                            data2.append(data[index])

                        elif len(data) == len(cells.mem_i):
                            for index in sorted(target_inds_mem, reverse=True):
                                del data[index]
                            data2.append(data[index])

                        # elif len(data) == len(cells.nn_i):
                        #     for index in sorted(target_inds_gj, reverse=True):
                        #         del data[index]
                        #     data2.append(data[index])

                    super_data2.append(data2)

                if type(super_data) == np.ndarray:
                    super_data2 = np.asarray(super_data2)

                setattr(sim, name, super_data2)


            else:
                data = getattr(sim, name)

                if isinstance(data, np.ndarray):
                    if len(data) == len(cells.cell_i):
                        data2 = np.delete(data, target_inds_cell)
                        setattr(sim, name, data2)

                    elif len(data) == len(cells.mem_i):
                        data2 = np.delete(data, target_inds_mem)
                        setattr(sim, name, data2)

                elif isinstance(data, list):
                    index = None
                    data2 = []

                    if len(data) == len(cells.cell_i):
                        for index in sorted(target_inds_cell, reverse=True):
                            del data[index]
                        data2.append(data[index])
                        setattr(sim, name, data2)

                    elif len(data) == len(cells.mem_i):
                        for index in sorted(target_inds_mem, reverse=True):
                            del data[index]
                        data2.append(data[index])
                        setattr(sim, name, data2)


        if p.Ca_dyn is True and sim.endo_retic is not None:

            sim.endo_retic.remove_ers(sim, target_inds_cell)


    #-------------------------------Fix-up cell world ----------------------------------------------------------------------
        new_cell_centres = []
        new_ecm_verts = []
        removal_flags = np.zeros(len(cells.cell_i))
        removal_flags[target_inds_cell] = 1

        for i,flag in enumerate(removal_flags):
            if flag == 0:
                new_cell_centres.append(cells.cell_centres[i])
                new_ecm_verts.append(cells.ecm_verts[i])

        cells.cell_centres = np.asarray(new_cell_centres)
        cells.ecm_verts = np.asarray(new_ecm_verts)

        # recalculate ecm_verts_unique:
        ecm_verts_flat,_,_ = tb.flatten(cells.ecm_verts)
        ecm_verts_set = set()

        for vert in ecm_verts_flat:
            ptx = vert[0]
            pty = vert[1]
            ecm_verts_set.add((ptx,pty))

        cells.ecm_verts_unique = [list(verts) for verts in list(ecm_verts_set)]
        cells.ecm_verts_unique = np.asarray(cells.ecm_verts_unique)  # convert to numpy array

        #-----------------------------------------------------------------
        logs.log_info('Recalculating cluster variables for new configuration...')

        cells.cellVerts(p)   # create individual cell polygon vertices and other essential data structures
        cells.cellMatrices(p)  # creates a variety of matrices used in routine cells calculations
        cells.intra_updater(p)  # creates matrix used for finite volume integration on cell patch
        cells.cell_vols(p)  # calculate the volume of cell and its internal regions
        cells.mem_processing(p)  # calculates membrane nearest neighbours, ecm interaction, boundary tags, etc
        cells.near_neigh(p)  # Calculate the nn array for each cell
        cells.voronoiGrid(p)
        cells.calc_gj_vects(p)
        cells.environment(p)  # define features of the ecm grid
        cells.make_maskM(p)
        cells.grid_len = len(cells.xypts)

        # update sim data lengths with new world values:
        sim.mdl = len(cells.mem_mids_flat)  # mems-data-length
        sim.cdl = len(cells.cell_centres)  # cells-data-length

        if p.is_ecm is True:  # set environnment data length
            sim.edl = len(cells.xypts)

        else:
            sim.edl = len(cells.mem_mids_flat)

        logs.log_info('Re-creating cell network Poisson solver...')
        cells.graphLaplacian(p)

        # if microtubules
        if sim.mtubes is not None:
            sim.mtubes.remove_mtubes(target_inds_mem, target_inds_cell, cells, sim, p)

        # if running voltage gated gap junctions, reinnitialize them:
        if p.v_sensitive_gj and sim.gj_funk is not None:
            sim.gj_funk.init(sim, cells, p)

        # delete data from molecules objects:
        if p.molecules_enabled and sim.molecules is not None:

            sim.molecules.core.mod_after_cut_event(target_inds_cell, target_inds_mem, sim, cells, p)

        if p.grn_enabled and sim.grn is not None:

            sim.grn.core.mod_after_cut_event(target_inds_cell, target_inds_mem, sim, cells, p)

        # Save target inds so they can be used outside of the core simulator (i.e. dynamically in sim-grn)
        sim.target_inds_cell_o = target_inds_cell
        sim.target_inds_mem_o = target_inds_mem

        # if hole_tag is False: # if we're not defining a hole at the beginning, reassign to new bflags
        sim.initDenv(cells,p)

        sim.conc_J_x = np.zeros(len(cells.xypts))
        sim.conc_J_y = np.zeros(len(cells.xypts))

        if p.fluid_flow or p.deformation:
            # make a laplacian and solver for discrete transfers on closed, irregular cell network:

            cells.deform_tools(p)

            if p.deformation:  # if user desires deformation:

                # create a copy of cells world, to apply deformations to for visualization purposes only:
                sim.cellso = copy.deepcopy(cells)

                if p.td_deform:
                    # make a laplacian and solver for discrete transfers on closed, irregular cell network
                    logs.log_info('Creating cell network Poisson solver...')
                    cells.graphLaplacian(p)

        # if p.sim_eosmosis is True:
        #
        #     if sim.move_pumps_channels is not None:
        #
        #         sim.move_pumps_channels.remove_data(target_inds_cell)

        # calculate targets for wound channel:
        match_inds = (sim.hurt_mask == 1.0).nonzero()

        mem_match_inds = tb.flatten(cells.cell_to_mems[match_inds[0]])[0]

        self.targets_vgWound = mem_match_inds

        if p.break_TJ: # we need to redo the TJ barrier at the wound site:

            # first identify indices of exterior and interior extracellular grid sites corresponding to the wound TJ:
            cell_wound_inds = (sim.hurt_mask == 1.0).nonzero()
            neigh_to_wound_cells, _, _ = tb.flatten(cells.cell_nn[cell_wound_inds])

            interior_wound_inds = cells.cell_to_mems[neigh_to_wound_cells]
            interior_wound_inds, _, _ = tb.flatten(interior_wound_inds)

            ecmtarg_a = cells.map_mem2ecm[self.targets_vgWound]
            ecmtarg_b = cells.map_mem2ecm[interior_wound_inds]

            ecmtarg = np.hstack((ecmtarg_a, ecmtarg_b))
            # ecmtarg = ecmtarg_a


            Dw = np.copy(sim.D_env_weight.ravel())

            # assign the wound TJ to the user's requested value:
            Dw[ecmtarg] = 1.0*p.wound_TJ

            # re-assign the environmental weight matrix:
            sim.D_env_weight = Dw.reshape(cells.X.shape)*1

            # furthermore, recalculate the individual ion diffusion maps:
            if p.is_ecm:

                for i, dmat in enumerate(sim.D_env):
                    Denv_o = np.copy(dmat)

                    Denv_o[ecmtarg] = sim.D_free[i]*p.wound_TJ*sim.Dtj_rel[i]

                    # re-assign the ecm diffusion grid filled with the environmental values at wound site:
                    sim.D_env[i] = Denv_o*1.0

            # # re-calculate specific maps of conductivity in cells and environment
            # sim.sigma_cell = np.asarray([((z ** 2) * p.q * p.F * cc * D) / (p.kb * p.T) for (z, cc, D) in
            #                               zip(sim.zs, sim.cc_cells, sim.D_free)]).mean(axis=0)
            #
            # sim.sigma_env = np.asarray(
            #     [((z ** 2) * p.q * p.F * cc * D) / (p.kb * p.T) for (z, cc, D) in
            #      zip(sim.zs, sim.cc_env, sim.D_env)]).mean(
            #     axis=0)
            #
            # if p.is_ecm:
            #
            #     # environmental conductivity matrix needs to be smoothed to assured simulation stability:
            #     sim.sigma_env = gaussian_filter(sim.sigma_env.reshape(cells.X.shape), 1).ravel()

        # need to also re-do tissue profiles and GJ
        self.tissueProfiles(sim, cells, p)
        cells.redo_gj(self, p)  # redo gap junctions to isolate different tissue types

        # If this is the fast BETSE solver, re-initialize this solver.
        if p.solver_type is SolverType.FAST:
            sim.fast_sim_init(cells, p)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **non-string sequence** (i.e., non-string object implementing the
abstract base class `collections.abc.Sequence`) facilities.

See Also
----------
betse.util.type.types.is_sequence
    Further details on what constitutes sequences and non-string sequences.
'''

#FIXME: Most if not all of the existing functionality provided by this submodule
#should be generalized to support higher-level iterable rather than lower-level
#sequence types and then shifted into the existing "iterables" submodule.
#Ideally, this submodule could then be excised entirely.

# ....................{ IMPORTS                            }....................
from betse.exceptions import BetseSequenceException
from betse.util.type import types
from betse.util.type.types import (
    type_check, CallableTypes, SequenceTypes, StrOrNoneTypes)
from collections.abc import Container, Mapping

# ....................{ EXCEPTIONS                         }....................
@type_check
def die_if_empty(
    *sequences: SequenceTypes, label: str = 'Sequence') -> None:
    '''
    Raise an exception prefixed by the passed label unless all passed sequences
    are **non-empty** (i.e., contain at least one element).

    Parameters
    ----------
    sequences: tuple
        Tuple of all sequences to be validated.
    label : optional[str]
        Human-readable label prefixing exception messages raised by this method.
        Defaults to a general-purpose string.

    Raises
    ----------
    BetseSequenceException
        If any passed sequence is empty.
    '''

    # If only one sequence is passed...
    if len(sequences) == 1:
        # If this sequence is non-empty, raise a simplistic exception.
        if is_empty(sequences[0]):
            raise BetseSequenceException('{} empty.'.format(label.capitalize()))
    # Else, multiple sequences are passed.
    else:
        # For each such sequence...
        for sequence_index, sequence in enumerate(sequences):
            # If this sequence is non-empty, raise an exception identifying the
            # index of this sequence in this parameter list.
            if is_empty(sequence):
                raise BetseSequenceException(
                    '{} {} empty.'.format(label.capitalize(), sequence_index))


@type_check
def die_unless_len(
    sequence: SequenceTypes,
    sequence_len: int,
    exception_message: StrOrNoneTypes = None,
) -> None:
    '''
    Raise an exception with the passed message (defaulting to a human-readable
    message) if the passed sequence is *not* of the passed length.

    Parameters
    ----------
    sequence: SequenceType
        Sequence to be validated.
    len: int
        Sequence length to test for.
    exception_message : optional[str]
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Raises
    ----------
    BetseSequenceException
        If this sequence is *not* of this length.
    '''

    # If this sequence is *NOT* of this length, raise an exception.
    if len(sequence) != sequence_len:
        # If no exception message was passed, synthesize one.
        if not exception_message:
            exception_message = 'Sequence length {} not {}.'.format(
                len(sequence), sequence_len)

        # Raise this exception.
        raise BetseSequenceException(exception_message)

# ....................{ TESTERS ~ type                     }....................
def is_sequence(*objs: object) -> bool:
    '''
    ``True`` only if all passed objects are **sequences** (i.e., of types
    conforming to but *not* necessarily subclassing the canonical
    :class:`collections.abc.Sequence` API).

    Parameters
    ----------
    objs: tuple[object]
        Tuple of all objects to be tested.

    Returns
    ----------
    bool
        ``True`` only if these objects are all sequences.

    See Also
    ----------
    :class:`collections.abc.Sequence`
        Further details.
    '''

    # all(). It is awesome.
    return all(isinstance(obj, SequenceTypes) for obj in objs)


def is_numpy_array(*objs: object) -> bool:
    '''
    ``True`` only if all passed objects are **Numpy arrays** (i.e., instances of
    the :class:`numpy.ndarray` superclass).

    Parameters
    ----------
    objs: tuple[object]
        Tuple of all objects to be tested.

    Returns
    ----------
    bool
        ``True`` only if these objects are all Numpy arrays.
    '''

    # Avoid importing third-party packages at the top level, for safety.
    from numpy import ndarray

    # all(). It exceeds at all the tests.
    return all(isinstance(obj, ndarray) for obj in objs)

# ....................{ TESTERS ~ len                      }....................
@type_check
def is_empty(*sequences: SequenceTypes) -> bool:
    '''
    ``True`` only if all passed sequences are **empty** (i.e., contain no
    elements).

    Parameters
    ----------
    sequences: tuple
        Tuple of all sequences to be tested.

    Returns
    ----------
    bool
        ``True`` only if these sequences are all empty.
    '''

    # all(). It is all awesome.
    #
    # To transparently support Numpy arrays, the length of this sequence *MUST*
    # be explicitly tested via the len() builtin rather than implicitly tested
    # as a boolean. On the latter, Numpy raises the following exception:
    #
    #     ValueError: The truth value of an array with more than one element is
    #     ambiguous. Use a.any() or a.all()
    return all(len(sequence) == 0 for sequence in sequences)

# ....................{ GETTERS                            }....................
#FIXME: Generalize into a function of the same name accepting an iterable rather
#than sequence and shift into the "iterables" submodule. When doing so, rename
#the "item_satisfier" parameter to "predicate" for orthogonality and clarity.
@type_check
def get_items_satisfying(
    sequence: SequenceTypes, item_satisfier: CallableTypes) -> SequenceTypes:
    '''
    New non-string sequence containing only the proper subset of elements from
    the passed non-string sequence satisfying the passed callable.

    This method effectively performs a general-purpose global search for
    sequence elements.

    Parameters
    ----------
    sequence : SequenceTypes
        Original sequence to return a proper subset of. For safety, this
        function does _not_ modify this sequence.
    item_satisfier : CallableTypes
        CallableTypes (e.g., function, lambda) accepting a single element of
        this sequence and returning only:
        * `True` if this element satisfies the desired requirements.
        * `False` otherwise.

    Returns
    ----------
    collections.Sequence
        Proper subset of the passed sequence. For efficiency, this new sequence
        is only a shallow rather than deep copy of the passed sequence.
    '''

    # Type of both the passed sequence and the sequence to be returned.
    sequence_type = type(sequence)

    # Return a generator-based shallow copy of this sequence.
    return sequence_type(item for item in sequence if item_satisfier(item))

# ....................{ GETTERS ~ str                      }....................
#FIXME: Generalize into a function of the same name accepting an iterable rather
#than sequence and shift into the "iterables" submodule. When doing so, rename
#the "item_prefix" parameter to simply "prefix".
@type_check
def get_items_prefixed_by(
    sequence: SequenceTypes, item_prefix: str) -> SequenceTypes:
    '''
    New non-string sequence containing only the proper subset of elements from
    the passed non-string sequence that are strings prefixed by the passed
    string prefix.

    This method effectively performs a prefix-based global search for sequence
    elements.

    Parameters
    ----------
    sequence : SequenceTypes
        Original sequence to return a proper subset of. For safety, this
        function does _not_ modify this sequence.
    item_prefix : str
        String prefixing all elements of the returned sequence.

    Returns
    ----------
    SequenceTypes
        Proper subset of the passed sequence. For efficiency, this new sequence
        is only a shallow rather than deep copy of the passed sequence.
    '''

    # Return a generator-based shallow copy of this sequence.
    return get_items_satisfying(
        sequence=sequence,
        item_satisfier=lambda item: (
            types.is_str(item) and item.startswith(item_prefix)
        ),
    )

# ....................{ OMITTERS                           }....................
def omit_item(sequence: SequenceTypes, item: object) -> SequenceTypes:
    '''
    New non-string sequence containing all elements of the first passed
    non-string sequence *not* equalling the passed object.

    Parameters
    ----------
    sequence : SequenceTypes
        Original sequence to return a proper subset of. For safety, this
        function does *not* modify this sequence.
    item : object
        Object to be omitted from the returned sequence.

    Returns
    ----------
    SequenceTypes
        Proper subset of the passed sequence. For efficiency, this new sequence
        is only a shallow rather than deep copy of the passed sequence.
    '''

    return omit_items(sequence=sequence, items=(item,))


@type_check
def omit_items(sequence: SequenceTypes, items: Container) -> SequenceTypes:
    '''
    New non-string sequence containing all elements of the passed non-string
    sequence *not* contained in the passed non-string container.

    Parameters
    ----------
    sequence : SequenceTypes
        Original sequence to return a proper subset of. For safety, this
        function does *not* modify this sequence.
    items : Container
        Container containing all elements to be omitted from the returned
        sequence.

    Returns
    ----------
    SequenceTypes
        Proper subset of the passed sequence. For efficiency, this new sequence
        is only a shallow rather than deep copy of the passed sequence.
    '''

    # Return a generator-based shallow copy of this sequence.
    return get_items_satisfying(
        sequence=sequence,
        item_satisfier=lambda item: item not in items,
    )

# ....................{ REMOVERS                           }....................
def remove_item(sequence: SequenceTypes, item: object) -> None:
    '''
    Remove all elements equal to the passed object from the passed mutable
    non-string sequence in-place.

    Parameters
    ----------
    sequence : collections.Sequence
        Sequence to remove elements from in-place.
    item : object
        Object to be removed.
    '''

    remove_items(sequence=sequence, items=(item,))


@type_check
def remove_items(sequence: SequenceTypes, items: Container) -> None:
    '''
    Remove all elements contained in the passed non-string container from
    the passed mutable non-string sequence in-place.

    Parameters
    ----------
    sequence : SequenceTypes
        Sequence to remove elements from in-place.
    items : Container
        Container containing all elements to be remove from this sequence.
    '''

    # Slice assignment implicitly modifies the original sequence, permitting
    # efficient reuse of existing assignment-based functionality.
    sequence[:] = omit_items(sequence, items)

# ....................{ REPLACERS                          }....................
@type_check
def replace_items(
    sequence: SequenceTypes, replacements: Mapping) -> SequenceTypes:
    '''
    New non-string sequence containing all elements of the passed non-string
    sequence (_in the same order_) such that each element equal to a key of the
    passed dictionary is replaced by the value of that key.

    This method effectively performs an equality-based global
    search-and-replacement for sequence elements.

    Parameters
    ----------
    sequence : collections.Sequence
        Original sequence to be returned transformed. For safety, this
        function does _not_ modify this sequence.
    replacements : collections.Mapping
        Mapping whose:
        * Keys are input values to find in the passed sequence. For simplicity,
          keys are matched via object equality rather than more complex object
          matching alternatives (e.g., glob, regex, prefix, suffix, substring).
        * Values are output values to replace these input values with in the
          returned sequence.

    Returns
    ----------
    collections.Sequence
        Sequence transformed from the passed sequence. For efficiency, this
        sequence is only a shallow rather than deep copy of the passed sequence.
    '''

    # Type of both the passed sequence and the sequence to be returned.
    sequence_type = type(sequence)

    # Return a shallow copy of this sequence, replacing each element that is a
    # key of the passed mapping by that key's value.
    return sequence_type(
        replacements[item] if item in replacements else item
        for item in sequence
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Simulation configuration file input and output (I/O) facilities.
'''

#FIXME: Validate the versions of loaded configuration files.

# ....................{ IMPORTS                            }....................
from betse import pathtree
from betse.util.io.log import logs
from betse.util.path import dirs, files, pathnames
from betse.util.path.dirs import DirOverwritePolicy
from betse.util.type.types import type_check  #, GeneratorType

# ....................{ WRITERS                            }....................
#FIXME: It should be feasible to replace this entire function (and hence remove
#this entire submodule) by:
#
#* Refactor the YamlFileABC.save() method signature to resemble this method's
#  signature as follows:
#
#    @type_check
#    def save(
#        # Mandatory parameters.
#        conf_filename: str,
#
#        # Optional parameters.
#        is_conf_file_overwritable: bool = False,
#        conf_subdir_overwrite_policy: DirOverwritePolicy = (
#            DirOverwritePolicy.SKIP_WITH_WARNING),
#    ) -> None:
#
#  Note our use of the "IGNORE_EXISTING" policy, which seems quite sensible,
#  safey, and sanitary for all YAML saving purposes.
#
#  *CAUTION*. Since the existing save() method performs overwriting by default,
#  we'll need to grep all calls to this method and explicitly pass the desired
#  parameters. Presumably, BETSEE already performs at least one such call.
#* Instantiating a "Parameters" object as follows:
#    p = Parameters().load(
#        conf_filename=pathtree.get_sim_config_default_filename())
#* Calling the save() method of this object.
#
#That's pretty obvious, frankly. Tragic that we didn't concoct it until now.
@type_check
def write_default(
    # Mandatory parameters.
    conf_filename: str,

    # Optional parameters.
    is_conf_overwritable: bool = False,
    is_data_overwritable: bool = False,
    # is_data_preservable: bool = False,
) -> None:
    '''
    Write a default YAML simulation configuration to the file with the passed
    path *and* recursively copy all external resources (e.g., images) required
    by this configuration into this file's parent directory.

    The resulting configuration will be usable as is with all high-level
    functionality requiring a valid configuration file (e.g., ``betse seed``).

    Parameters
    ----------
    conf_filename : str
        Absolute or relative path of the target YAML file to be written.
    is_conf_overwritable : optional[bool]
        ``True`` if an existing target YAML file is to be silently overwritten
        *or* ``False`` if an exception is to be raised if this file already
        exists. Defaults to ``False``.
    is_data_overwritable : optional[bool]
        ``True`` if existing target resources required by this target YAML file
        are to be silently overwritten *or* ``False`` if an exception is to be
        raised if any such resource already exists. Defaults to ``False``.
    is_data_preservable : optional[bool]
        ``True`` if existing target resources required by this target YAML file
        are to be preserved "as is" with a non-fatal warning *or* ``False`` if
        an exception is to be raised if any such resource already exists.
        Defaults to ``False``. The ``is_data_overwritable`` parameter takes
        precedence over this parameter. Specifically, if the
        ``is_data_overwritable`` parameter is ``True`` and this parameter is:
        * ``True``, an exception is raised. These two parameters conflict and
          hence *cannot* both be concurrently enabled.
        * ``False``, this parameter is silently ignored.

    Raises
    ----------
    BetseFileException
        If this file already exists *and* ``is_overwritable`` is ``False``.
    '''

    # Announce the ugly shape of things to come.
    logs.log_info('Writing default simulation configuration...')

    # Copy the default source simulation configuration file to this target file.
    files.copy(
        src_filename=pathtree.get_sim_config_default_filename(),
        trg_filename=conf_filename,
        is_overwritable=is_conf_overwritable,
    )

    # Source directory containing the default simulation configuration.
    src_dirname = pathtree.get_data_yaml_dirname()

    # Target directory to be copied into, defined to be either the parent
    # directory of the passed path if this path has a dirname or the current
    # working directory otherwise.
    trg_dirname = pathnames.get_dirname_or_cwd(conf_filename)

    # Note that the simple solution of recursively copying this source directory
    # into the parent directory of the passed target file (e.g., by calling
    # "dirs.copy(src_dirname, pathnames.get_dirname(conf_filename))") fails
    # for the following subtle reasons:
    #
    # * This target directory may be already exist, which dirs.copy() prohibits
    #   even if the directory is empty.
    # * This target configuration file basename may differ from that of the
    #   default source configuration file basename, necessitating a subsequent
    #   call to file.move().
    #
    # For each subdirectory of this directory...
    for src_subdirname in dirs.iter_subdirnames(src_dirname):
        # Recursively copy from this subdirectory into the target directory.
        dirs.copy_into_dir(
            src_dirname=src_subdirname,
            trg_dirname=trg_dirname,

            # Ignore (i.e., skip) each target resource of this subdirectory that
            # already exists with a non-fatal warning. This is purely a caller
            # convenience, permitting multiple configuration files with
            # different basenames to be created in the same parent directory
            # *WITHOUT* either raising exceptions on or silently overwriting
            # duplicate target resources.
            overwrite_policy=DirOverwritePolicy.SKIP_WITH_WARNING,

            # Ignore all empty ".gitignore" files in all subdirectories of this
            # source directory. These files serve as placeholders instructing
            # Git to track their parent subdirectories, but otherwise serve no
            # purpose. Preserving these files only invites end user confusion.
            ignore_basename_globs=('.gitignore',),
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Collection of the absolute paths of numerous critical files and directories
describing the structure of this application on the local filesystem.

These are intended for consumption by both this application and downstream
reverse dependencies of this application (e.g., BETSEE, the BETSE GUI). For
portability, these paths are initialized in a system-aware manner guaranteed to
be sane under insane installation environments -- including PyInstaller-frozen
executables and :mod:`setuptools`-installed script wrappers.
'''

#FIXME: The current globals-based approach is inefficient in the case of BETSE
#being installed as a compressed EGG rather than an uncompressed directory. In
#the former case, the current approach (namely, the call to
#resources.get_pathname() performed below) silently extracts the entirety of
#this egg to a temporary setuptools-specific cache directory. That's bad. To
#circumvent this, we'll need to refactor the codebase to directly require only
#"file"-like objects rather than indirectly requiring the absolute paths of
#data resources that are then opened as "file"-like objects.
#
#Specifically, whenever we require a "file"-like object for a codebase resource,
#we'll need to call the setuptools-specific pkg_resources.resource_stream()
#function rather than attempting to open the path given by a global below.
#Ultimately, *ALL* of the codebase-specific globals declared below (e.g.,
#"DATA_DIRNAME") should go away.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on missing mandatory dependencies,
# the top-level of this module may import *ONLY* from packages guaranteed to
# exist at installation time (i.e., standard Python packages). Likewise, to
# avoid circular import dependencies, the top-level of this module should avoid
# importing application packages except where explicitly required.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
import os
from betse import metadata
from betse.exceptions import BetseModuleException
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import type_check, StrOrNoneTypes

# ....................{ GETTERS ~ dir                      }....................
@func_cached
def get_root_dirname() -> str:
    '''
    Absolute pathname of the root directory suffixed by a directory separator.

    The definition of "root directory" conditionally depends on the current
    platform. If this platform is:

    * POSIX-compatible (e.g., Linux, OS X), this is simply ``/``.
    * Microsoft Windows, this is the value of the ``%HOMEDRIVE%`` environment
      variable. This is the ``:``-suffixed letter of the drive to which Windows
      was originally installed -- typically but *not* necessarily ``C:\\``.
    '''

    # Avoid circular import dependencies.
    from betse.util.os import oses
    from betse.util.os.shell import shellenv

    # Return this dirname.
    if oses.is_windows_vanilla():
        return shellenv.get_var_or_default('HOMEDRIVE', 'C:') + os.path.sep
    else:
        return os.path.sep


@func_cached
def get_home_dirname() -> str:
    '''
    Absolute pathname of the home directory of the current user if found *or*
    raise an exception otherwise (i.e., if this directory is *not* found).
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs, pathnames

    # Absolute path of this directory.
    home_dirname = pathnames.canonicalize('~')

    # If this directory is not found, fail.
    dirs.die_unless_dir(home_dirname)

    # Return this directory's path.
    return home_dirname

# ....................{ GETTERS ~ dir : app                }....................
@func_cached
def get_dot_dirname() -> str:
    '''
    Absolute pathname of this application's top-level dot directory in the home
    directory of the current user, silently creating this directory if *not*
    already found.

    This directory contains user-specific files (e.g., logfiles, profile files)
    both read from and written to at application runtime. These are typically
    plaintext files consumable by external users and third-party utilities.

    Locations
    ----------
    Specifically, this path is:

    * Under Linux, ``~/.betse/``. BETSE does *not* currently comply with the
      _XDG Base Directory Specification (e.g., ``~/.local/share/betse``), which
      the principal authors of BETSE regard as unuseful (if not harmful).
    * Under OS X, ``~/Library/Application Support/betse``.
    * Under Windows,
      ``C:\\Documents and Settings\\${User}\\Application Data\\betse``.

    .. _XDG Base Directory Specification:
        http://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html

    Raises
    ----------
    BetseDirException
        If this directory does *not* exist.
    '''

    # Avoid circular import dependencies.
    from betse.util.os import oses
    from betse.util.os.shell import shellenv
    from betse.util.path import dirs, pathnames

    # Absolute path of this directory.
    dot_dirname = None

    # If the current platform is macOS, return the appropriate directory.
    if oses.is_macos():
        dot_dirname = pathnames.join(
            get_home_dirname(),
            'Library',
            'Application Support',
            metadata.SCRIPT_BASENAME,
        )
    # If the current platform is Windows, return the appropriate directory.
    elif oses.is_windows():
        dot_dirname = pathnames.join(shellenv.get_var('APPDATA'), metadata.NAME)
    # Else, assume the current platform to be POSIX-compatible.
    else:
        #FIXME: Explicitly assert POSIX compatibility here. To do so, we'll want
        #to define and call a new betse.util.os.oses.die_unless_posix()
        #function here.

        dot_dirname = pathnames.join(
            get_home_dirname(), '.' + metadata.SCRIPT_BASENAME)

    # Create this directory if not found.
    dirs.make_unless_dir(dot_dirname)

    # Return this directory's path.
    return dot_dirname

# ....................{ GETTERS ~ dir : data               }....................
@func_cached
def get_data_dirname() -> str:
    '''
    Absolute pathname of this application's top-level data directory if found
    *or* raise an exception otherwise (i.e., if this directory is *not* found).

    This directory contains application-internal resources (e.g., media files)
    required at application runtime.

    Raises
    ----------
    BetseDirException
        If this directory does *not* exist.
    '''

    # Avoid circular import dependencies.
    import betse
    from betse.util.path import dirs, pathnames

    # Absolute path of this directory.
    data_dirname = pathnames.get_app_pathname(package=betse, pathname='data')

    # If this directory is not found, raise an exception.
    dirs.die_unless_dir(data_dirname)

    # Return the absolute path of this directory.
    return data_dirname


@func_cached
def get_data_yaml_dirname() -> str:
    '''
    Absolute pathname of this application's data subdirectory containing
    YAML-formatted files if found *or* raise an exception otherwise (i.e., if
    this directory is *not* found).

    This directory contains:

    * The YAML-formatted plaintext file specifying a simulation configuration.
    * All assets referenced in and hence required by this file.

    Raises
    ----------
    BetseDirException
        If this directory does *not* exist.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs

    # Return this dirname if this directory exists or raise an exception.
    return dirs.join_and_die_unless_dir(get_data_dirname(), 'yaml')

# ....................{ GETTERS ~ dir : package            }....................
@func_cached
def get_package_dirname() -> str:
    '''
    Absolute pathname of this application's top-level package directory if found
    *or* raise an exception otherwise (i.e., if this directory is *not* found).

    This directory typically resides in the ``site-packages`` subdirectory of
    the system-wide standard lib for the active Python interpreter (e.g.,
    ``/usr/lib64/python3.6/site-packages/betse``).

    Raises
    ----------
    BetseDirException
        If this directory does *not* exist.
    '''

    # Avoid circular import dependencies.
    import betse
    from betse.util.path import dirs
    from betse.util.type import modules

    # Absolute pathname of the directory yielding the top-level "betse" package.
    package_dirname = modules.get_dirname(betse)

    # If this directory is not found, fail.
    dirs.die_unless_dir(package_dirname)

    # Return this directory's pathname.
    return package_dirname

# ....................{ GETTERS ~ dir : git                }....................
@func_cached
def get_git_worktree_dirname() -> str:
    '''
    Absolute pathname of this application's Git-based **working tree** (i.e.,
    top-level directory containing this application's ``.git`` subdirectory and
    ``setup.py`` install script) if this application was installed in a
    developer manner *or* raise an exception otherwise (i.e., if this directory
    is *not* found).

    Raises
    ----------
    BetseDirException
        If this directory does *not* exist.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs

    # Absolute pathname of this application's Git-based working tree if this
    # application was installed in a developer manner or "None" otherwise.
    git_worktree_dirname = get_git_worktree_dirname_or_none()

    # If this directory is not found, fail.
    dirs.die_unless_dir(git_worktree_dirname)

    # Return this directory's pathname.
    return git_worktree_dirname


@func_cached
def get_git_worktree_dirname_or_none() -> StrOrNoneTypes:
    '''
    Absolute pathname of this application's Git-based **working tree** (i.e.,
    top-level directory containing this application's ``.git`` subdirectory and
    ``setup.py`` install script) if this application was installed in a
    developer manner *or* ``None`` otherwise.

    Returns
    ----------
    StrOrNoneTypes
        Specifically, this function returns either:
        * A pathname if this application was installed in a developer manner,
          typically either via:
          * ``python3 setup.py develop``.
          * ``python3 setup.py symlink``.
        * ``None`` if this application was installed in a non-developer manner,
          typically either via:
          * ``pip3 install``.
          * ``python3 setup.py develop``.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs, pathnames

    # Absolute pathname of the directory offering the top-level "betse" package,
    # canonicalized into a directory rather than symbolic link to increase the
    # likelihood of obtaining the actual parent directory of this package.
    package_dirname = pathnames.canonicalize(get_package_dirname())

    # Absolute pathname of the parent directory of this directory.
    worktree_dirname = pathnames.get_dirname(package_dirname)

    # Absolute pathname of the ".git" subdirectory of this parent directory.
    git_subdirname = pathnames.join(worktree_dirname, '.git')

    # Return this parent directory's absolute pathname if this subdirectory
    # exists *OR* "None" otherwise.
    return worktree_dirname if dirs.is_dir(git_subdirname) else None

# ....................{ GETTERS ~ file                     }....................
@func_cached
def get_log_default_filename() -> str:
    '''
    Absolute pathname of this application's default user-specific logfile.

    This is the plaintext file to which all messages are logged by default.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # Return the absolute path of this file.
    return pathnames.join(get_dot_dirname(), metadata.SCRIPT_BASENAME + '.log')


@func_cached
def get_profile_default_filename() -> str:
    '''
    Absolute pathname of this application's default user-specific profile
    dumpfile.

    This is the binary file to which profiled statistics are saved by default.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # Return the absolute path of this file.
    return pathnames.join(get_dot_dirname(), metadata.SCRIPT_BASENAME + '.prof')


@func_cached
def get_sim_config_default_filename() -> str:
    '''
    Absolute pathname of this application's default simulation configuration
    file if found *or* raise an exception otherwise (i.e., if this file is *not*
    found).

    This is the plaintext YAML-formatted file from which all new simulation
    configurations derive.

    Raises
    ----------
    BetseFileException
        If this file does *not* exist.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import files

    # Return this dirname if this directory exists or raise an exception.
    return files.join_and_die_unless_file(
        get_data_yaml_dirname(), 'sim_config.yaml')

# ....................{ GETTERS ~ file : arg               }....................
@type_check
def get_repl_history_filename(repl_module_name: str) -> dict:
    '''
    Absolute pathname of this application's default user-specific history file
    for the read–eval–print loop (REPL) implemented by the third-party module
    with the passed name.

    This history file is a plaintext file to which this REPL appends each read
    user command for preserving command history between REPL sessions.

    Parameters
    ----------
    repl_module_name : str
        Fully-qualified name of the third-party module implementing this REPL.

    Returns
    ----------
    str
        Absolute path of the user-specific history file for this REPL.

    Raises
    ----------
    BetseModuleException
        If this REPL module name is unrecognized.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # Absolute path of this application's user-specific dot directory.
    DOT_DIRNAME = get_dot_dirname()

    # Dictionary mapping each REPL module name to its history filename.
    REPL_MODULE_NAME_TO_HISTORY_FILENAME = {
        'ptpython': pathnames.join(DOT_DIRNAME, 'ptpython.hist'),
        'readline': pathnames.join(DOT_DIRNAME, 'readline.hist'),
    }

    # If this REPL module name is unrecognized, fail.
    if repl_module_name not in REPL_MODULE_NAME_TO_HISTORY_FILENAME:
        raise BetseModuleException(
            'REPL module "{}" unrecognized.'.format(repl_module_name))

    # Else, return the history filename for this REPL.
    return REPL_MODULE_NAME_TO_HISTORY_FILENAME[repl_module_name]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Global unit test configuration for BETSE.

`py.test` implicitly imports all functionality defined by this module into all
unit test modules. As this functionality includes all publicly declared
functional fixtures in this `fixture` subpackage, these tests may reference
these fixtures without explicit imports.
'''

# ....................{ IMPORTS ~ fixture                  }....................



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **reference** (i.e., pointer to the address of an object) facilities.
'''

# ....................{ IMPORTS                            }....................
import weakref
from betse.util.type.types import WeakRefProxyTypes

# ....................{ GETTERS                            }....................
def proxy_weak(obj: object) -> WeakRefProxyTypes:
    '''
    Weak reference to the passed object as a proxy object, such that the latter
    transparently masquerades as the former.

    If the passed object is:

    * Already a proxy object, this object is returned as is.
    * Callable, a :class:`CallableProxyType` object proxying this object is
      returned.
    * Uncallable, a :class:`ProxyType` object proxying this object is returned.

    This function should *always* be called in lieu of the lower-level
    :func:`weakref.proxy` function, which unsafely raises the following
    exception when passed an existing proxy object:

        TypeError: cannot create weak reference to 'weakproxy' object

    Parameters
    ----------
    obj : object
        Object to be proxied.

    Returns
    ----------
    WeakRefProxyTypes
        Weak reference to the passed object as a proxy object. When the passed
        object is garbage collected, this proxy object yields ``None`` rather
        than the passed object. This proxy object is unhashable regardless of
        the hashability of the passed object, minimizing issues arising from the
        fundamentally mutable nature of weak references and preventing their use
        as dictionary keys.
    '''

    # If this object is already a weak reference proxy, return the object as is.
    if isinstance(obj, WeakRefProxyTypes):
        return obj
    # Else, return a weak reference proxy to the object.
    else:
        return weakref.proxy(obj)
